"""
@author: Anthony Soulain (University of Sydney)

-------------------------------------------------------------------------
AMICAL: Aperture Masking Interferometry Calibration and Analysis Library
-------------------------------------------------------------------------

OIFITS related function.

--------------------------------------------------------------------
"""

import datetime
import os

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time
from astroquery.simbad import Simbad
from matplotlib import pyplot as plt
from munch import munchify as dict2class
from termcolor import cprint

from amical.tools import rad2mas

list_color = ["#00a7b5", "#afd1de", "#055c63", "#ce0058", "#8a8d8f", "#f1b2dc"]


def _compute_flag(value, sigma, limit=4.0):
    """Compute flag array using snr (snr < 4 by default)."""
    npts = len(value)
    flag = np.array([False] * npts)
    snr = abs(value / sigma)
    cond = snr <= limit
    flag[cond] = True
    return flag


def _peak1hole_cp(bs, ihole=0):
    """Get the indices of each CP including the given hole."""
    bs2bl_ix = bs.mask.bs2bl_ix
    bl2h_ix = bs.mask.bl2h_ix
    sel_ind = []
    for i, x in enumerate(bs2bl_ix.T):
        cpi = [bl2h_ix.T[y] for y in x]
        if (ihole in cpi[0]) or (ihole in cpi[1]) or (ihole in cpi[2]):
            sel_ind.append(i)
    sel_ind = np.array(sel_ind)
    return sel_ind


def _format_staindex_v2(tab):
    """Format the sta_index of v2 to save as oifits."""
    sta_index = []
    for x in tab:
        ap1 = int(x[0])
        ap2 = int(x[1])
        line = np.array([ap1, ap2]) + 1
        sta_index.append(line)
    return sta_index


def _format_staindex_t3(tab):
    """Format the sta_index of cp to save as oifits."""
    sta_index = []
    for x in tab:
        ap1 = int(x[0])
        ap2 = int(x[1])
        ap3 = int(x[2])
        line = np.array([ap1, ap2, ap3]) + 1
        sta_index.append(line)
    return sta_index


def _apply_flag(dict_calibrated, unit="arcsec"):
    """Apply flag and convert to appropriete units."""

    wl = dict_calibrated["OI_WAVELENGTH"]["EFF_WAVE"]
    uv_scale = {
        "m": 1,
        "rad": 1 / wl,
        "arcsec": 1 / wl / rad2mas(1e-3),
        "lambda": 1 / wl / 1e6,
    }

    U = dict_calibrated["OI_VIS2"]["UCOORD"] * uv_scale[unit]
    V = dict_calibrated["OI_VIS2"]["VCOORD"] * uv_scale[unit]

    flag_v2 = np.invert(dict_calibrated["OI_VIS2"]["FLAG"])
    V2 = dict_calibrated["OI_VIS2"]["VIS2DATA"][flag_v2]
    e_V2 = dict_calibrated["OI_VIS2"]["VIS2ERR"][flag_v2] * 1
    sp_freq_vis = dict_calibrated["OI_VIS2"]["BL"][flag_v2] * uv_scale[unit]
    flag_cp = np.invert(dict_calibrated["OI_T3"]["FLAG"])
    cp = dict_calibrated["OI_T3"]["T3PHI"][flag_cp]
    e_cp = dict_calibrated["OI_T3"]["T3PHIERR"][flag_cp]
    sp_freq_cp = dict_calibrated["OI_T3"]["BL"][flag_cp] * uv_scale[unit]
    bmax = 1.2 * np.max(np.sqrt(U ** 2 + V ** 2))

    cal_flagged = dict2class(
        {
            "U": U,
            "V": V,
            "bmax": bmax,
            "vis2": V2,
            "e_vis2": e_V2,
            "cp": cp,
            "e_cp": e_cp,
            "sp_freq_vis": sp_freq_vis,
            "sp_freq_cp": sp_freq_cp,
            "wl": wl[0],
            "band": dict_calibrated["info"]["FILT"],
        }
    )

    return cal_flagged


def cal2dict(
    cal,
    target=None,
    pa=0,
    del_pa=0,
    snr=4,
    true_flag_v2=True,
    true_flag_t3=False,
    include_vis=False,
    oriented=-1,
    ind_hole=None,
):
    """Format class containing calibrated data into appropriate dictionnary
    format to be saved as oifits files.

    Parameters
    ----------
    `cal` : {class}
        Class returned by amical.calibrate function (see core.py),\n
    `target` : {str}, (optional),
        Name of the science target, by default '',\n
    `fake_obj` : {bool}, (optional),
        If True, observables extracted from simulated data (celestial
        coordinates are omitted), by default False,\n
    `pa` : {int}, (optional)
        Position angle, by default 0 [deg]\n
    `del_pa` : {int}, (optional)
        Uncertainties of parallactic angle , by default 0\n
    `true_flag_v2` : {bool}, (optional)
        If True, real flag are computed for v2 using snr threshold
        (default 4), by default True,\n
    `true_flag_t3` : {bool}, (optional)
        If True, real flag are computed for cp using snr threshold
        (default 4), by default True,\n
    `include_vis` : {bool}, (optional)
        If True, include visibility amplitude in the oifits dictionnary,
         by default False,\n
    `oriented` {float}:
        If oriented == -1, east assumed to the left in the image, otherwise
        oriented == 1 (east to the right); (Default -1),\n

    Returns
    -------
    `dic`: {dict}
        Dictionnary format of the data to be save as oifits.
    """
    res_t = cal.raw_t
    res_c = cal.raw_c
    n_baselines = res_t.mask.n_baselines
    n_bs = len(cal.cp)
    bl2h_ix = res_t.mask.bl2h_ix
    bs2bl_ix = res_t.mask.bs2bl_ix

    date = "2020-02-07T00:54:11"
    exp_time = 0.8
    try:
        ins = res_t.infos.instrument
        maskname = res_t.infos.maskname
        pixscale = res_t.infos.pixscale
    except KeyError:
        cprint(
            "Error: 'INSTRUME', 'NRMNAME' or 'PIXELSCL' are not in the header.", "red"
        )
        return None

    t = Time(date, format="isot", scale="utc")

    if true_flag_v2:
        flagV2 = _compute_flag(cal.vis2, cal.e_vis2, limit=snr)
    else:
        flagV2 = np.array([False] * n_baselines)

    if true_flag_t3:
        flagCP = _compute_flag(cal.cp, cal.e_cp, limit=snr)
    else:
        flagCP = np.array([False] * n_bs)

    sta_index_v2 = []
    for i in range(n_baselines):
        sta_index_v2.append(np.array(bl2h_ix[:, i]))
    sta_index_v2 = np.array(sta_index_v2)

    if target is None:
        target = res_t.infos.target

    thepa = pa - 0.5 * del_pa
    u, v = oriented * res_t.u, res_t.v
    u1 = u * np.cos(np.deg2rad(thepa)) + v * np.sin(np.deg2rad(thepa))
    v1 = -u * np.sin(np.deg2rad(thepa)) + v * np.cos(np.deg2rad(thepa))

    if type(res_c) is list:
        calib_name = res_c[0].infos.target
    else:
        calib_name = res_c.infos.target

    if ind_hole is not None:
        cprint("Select only independant CP using common hole #%i." % ind_hole, "green")
        sel_ind_cp = _peak1hole_cp(cal.raw_t, ind_hole)
    else:
        sel_ind_cp = np.arange(len(cal.cp))

    dic = {
        "OI_VIS2": {
            "VIS2DATA": cal.vis2,
            "VIS2ERR": cal.e_vis2,
            "UCOORD": u1,
            "VCOORD": v1,
            "STA_INDEX": sta_index_v2,
            "MJD": t.mjd,
            "INT_TIME": exp_time,
            "TIME": 0,
            "TARGET_ID": 1,
            "FLAG": flagV2,
            "BL": (u1 ** 2 + v1 ** 2) ** 0.5,
        },
        "OI_T3": {
            "MJD": t.mjd,
            "INT_TIME": exp_time,
            "T3PHI": cal.cp[sel_ind_cp],
            "T3PHIERR": cal.e_cp[sel_ind_cp],
            "TIME": 0,
            "T3AMP": np.zeros(len(cal.cp))[sel_ind_cp],
            "T3AMPERR": np.zeros(len(cal.cp))[sel_ind_cp],
            "U1COORD": u1[bs2bl_ix[0, :]][sel_ind_cp],
            "V1COORD": v1[bs2bl_ix[0, :]][sel_ind_cp],
            "U2COORD": u1[bs2bl_ix[1, :]][sel_ind_cp],
            "V2COORD": v1[bs2bl_ix[1, :]][sel_ind_cp],
            "STA_INDEX": list(np.array(res_t.mask.closing_tri)[sel_ind_cp]),
            "FLAG": flagCP[sel_ind_cp],
            "BL": res_t.bl_cp[sel_ind_cp],
        },
        "OI_WAVELENGTH": {
            "EFF_WAVE": np.array([res_t.wl]),
            "EFF_BAND": np.array([res_t.e_wl]),
        },
        "info": {
            "TARGET": target,
            "CALIB": calib_name,
            "FILT": res_t.infos.filtname,
            "INSTRUME": ins,
            "MASK": maskname,
            "MJD": t.mjd,
            "HDR": res_t.infos,
            "ISZ": res_t.infos.isz,
            "PSCALE": pixscale,
            "xycoord": res_t.mask.xycoord,
            "SEEING": res_t.infos.seeing,
        },
    }

    if include_vis:
        dic["OI_VIS"] = {
            "TARGET_ID": 1,
            "TIME": 0,
            "MJD": t.mjd,
            "INT_TIME": exp_time,
            "VISAMP": cal.visamp,
            "VISAMPERR": cal.e_visamp,
            "VISPHI": cal.visphi,
            "VISPHIERR": cal.e_visphi,
            "UCOORD": u1,
            "VCOORD": v1,
            "STA_INDEX": sta_index_v2,
            "FLAG": flagV2,
        }
    return dic


def load(
    filename,
    filtname=None,
):
    """Load an oifits file format and store it as dictionnary. The different keys are
    representative of the oifits standard structure ('OI_WAVELENGTH', 'OI_VIS2', etc.)

    Parameters
    ----------
    filename {str}:
        Name of the oifits file,\n
    filtname {str}:
        Name of the filter used if not included in the header (default: None)

    Output:
    -------
    dic {dict}:
        Dictionnary containing the oifits file data.
    """
    fitsHandler = fits.open(filename)
    hdr = fitsHandler[0].header

    dic = {}
    for hdu in fitsHandler[1:]:
        if hdu.header["EXTNAME"] == "OI_WAVELENGTH":
            dic["OI_WAVELENGTH"] = {
                "EFF_WAVE": hdu.data["EFF_WAVE"],
                "EFF_BAND": hdu.data["EFF_BAND"],
            }

        if hdu.header["EXTNAME"] == "OI_VIS2":
            dic["OI_VIS2"] = {
                "VIS2DATA": hdu.data["VIS2DATA"],
                "VIS2ERR": hdu.data["VIS2ERR"],
                "UCOORD": hdu.data["UCOORD"],
                "VCOORD": hdu.data["VCOORD"],
                "STA_INDEX": hdu.data["STA_INDEX"],
                "MJD": hdu.data["MJD"],
                "INT_TIME": hdu.data["INT_TIME"],
                "TIME": hdu.data["TIME"],
                "TARGET_ID": hdu.data["TARGET_ID"],
                "FLAG": np.array(hdu.data["FLAG"]),
            }
            try:
                dic["OI_VIS2"]["BL"] = hdu.data["BL"]
            except KeyError:
                dic["OI_VIS2"]["BL"] = (
                    hdu.data["UCOORD"] ** 2 + hdu.data["VCOORD"] ** 2
                ) ** 0.5

        if hdu.header["EXTNAME"] == "OI_VIS":
            dic["OI_VIS"] = {
                "TARGET_ID": hdu.data["TARGET_ID"],
                "TIME": hdu.data["TIME"],
                "MJD": hdu.data["MJD"],
                "INT_TIME": hdu.data["INT_TIME"],
                "VISAMP": hdu.data["VISAMP"],
                "VISAMPERR": hdu.data["VISAMPERR"],
                "VISPHI": hdu.data["VISPHI"],
                "VISPHIERR": hdu.data["VISPHIERR"],
                "UCOORD": hdu.data["UCOORD"],
                "VCOORD": hdu.data["VCOORD"],
                "STA_INDEX": hdu.data["STA_INDEX"],
                "FLAG": hdu.data["FLAG"],
            }
            try:
                dic["OI_VIS"]["BL"] = hdu.data["BL"]
            except KeyError:
                dic["OI_VIS"]["BL"] = (
                    hdu.data["UCOORD"] ** 2 + hdu.data["VCOORD"] ** 2
                ) ** 0.5

        if hdu.header["EXTNAME"] == "OI_T3":
            u1 = hdu.data["U1COORD"]
            u2 = hdu.data["U2COORD"]
            v1 = hdu.data["V1COORD"]
            v2 = hdu.data["V2COORD"]
            u3 = -(u1 + u2)
            v3 = -(v1 + v2)
            bl_cp = []
            for k in range(len(u1)):
                B1 = np.sqrt(u1[k] ** 2 + v1[k] ** 2)
                B2 = np.sqrt(u2[k] ** 2 + v2[k] ** 2)
                B3 = np.sqrt(u3[k] ** 2 + v3[k] ** 2)
                bl_cp.append(np.max([B1, B2, B3]))  # rad-1
            bl_cp = np.array(bl_cp)

            dic["OI_T3"] = {
                "T3PHI": hdu.data["T3PHI"],
                "T3PHIERR": hdu.data["T3PHIERR"],
                "T3AMP": hdu.data["T3AMP"],
                "T3AMPERR": hdu.data["T3AMPERR"],
                "U1COORD": hdu.data["U1COORD"],
                "V1COORD": hdu.data["V1COORD"],
                "U2COORD": hdu.data["U2COORD"],
                "V2COORD": hdu.data["V2COORD"],
                "STA_INDEX": hdu.data["STA_INDEX"],
                "MJD": hdu.data["MJD"],
                "FLAG": hdu.data["FLAG"],
                "TARGET_ID": hdu.data["TARGET_ID"],
                "TIME": hdu.data["TIME"],
                "INT_TIME": hdu.data["INT_TIME"],
            }
            try:
                dic["OI_T3"]["BL"] = hdu.data["FREQ"]
            except KeyError:
                dic["OI_T3"]["BL"] = bl_cp

    dic["info"] = {h: hdr[h] for h in hdr}  # {'MJD': mjd,

    if "FILT" not in list(dic["info"].keys()):
        if filtname is None:
            print("No filter in info or as input param?")
        else:
            dic["info"]["FILT"] = filtname
    return dic


def loadc(filename):
    """Same as load but provide an easy usable output as a class format (output.v2, or output.cp)."""
    dic = load(filename)
    res = {}
    # Extract infos
    res["target"] = dic["info"].get("OBJECT")

    res["calib"] = dic["info"].get("CALIB")
    res["seeing"] = dic["info"].get("SEEING")
    res["mjd"] = dic["info"].get("MJD")
    if res["mjd"] is None:
        res["mjd"] = dic["info"].get("MJD-OBS")

    # Extract wavelength
    res["wl"] = dic["OI_WAVELENGTH"]["EFF_WAVE"]
    res["e_wl"] = dic["OI_WAVELENGTH"]["EFF_BAND"]

    # Extract squared visibilities
    res["vis2"] = dic["OI_VIS2"]["VIS2DATA"]
    res["e_vis2"] = dic["OI_VIS2"]["VIS2ERR"]
    res["u"] = dic["OI_VIS2"]["UCOORD"]
    res["v"] = dic["OI_VIS2"]["VCOORD"]
    res["bl"] = dic["OI_VIS2"]["BL"]
    res["flag_vis"] = dic["OI_VIS2"]["FLAG"]

    # Extract closure phases
    res["cp"] = dic["OI_T3"]["T3PHI"]
    res["e_cp"] = dic["OI_T3"]["T3PHIERR"]
    res["u1"] = dic["OI_T3"]["U1COORD"]
    res["v1"] = dic["OI_T3"]["V1COORD"]
    res["u2"] = dic["OI_T3"]["U2COORD"]
    res["v2"] = dic["OI_T3"]["V2COORD"]
    res["bl_cp"] = dic["OI_T3"]["BL"]
    res["flag_cp"] = dic["OI_T3"]["FLAG"]

    return dict2class(res)


def save(
    input_calibrated,
    oifits_file=None,
    datadir="Saveoifits",
    pa=0,
    ind_hole=None,
    fake_obj=False,
    include_vis=False,
    true_flag_v2=True,
    true_flag_t3=False,
    snr=4,
    verbose=False,
):
    """
    Summary:
    --------

    Save the class object (from calibrate function) into oifits format. The input
    input_calibrated can be a list of object for IFU data (e.g.: IFS-SPHERE).

    Parameters:
    -----------

    `input_calibrated` {class}:
        Class or list of class containing all calibrated interferometric variable extracted
        using calibrate (amical.calibration) function,\n
    `ind_hole` {int}:
        By default, ind_hole is None, all the CP are considered ncp = N(N-1)(N-2)/6. If
        ind_hole is set, save only the independant CP including the given hole
        ncp = (N-1)(N-2)/2.\n
    `oifits_file` {str}:
        Name of the oifits file,\n
    `include_vis` {bool}:
        If True, include OI_VIS table in the oifits,\n
    `fake_obj` {bool}:
        If True, observable are extracted from simulated data and so doesn't
        contain real target informations (simbad search is ignored),\n
    `pa` {float}:
        Position angle of the observation (i.e.: north direction) [deg],\n
    `true_flag_v2`, `true_flag_t3` {bool}:
        if True, the true flag are used using snr,\n
    `snr` {float}:
        Limit snr used to compute flags (default=4),\n
    `datadir` {str}:
        Folder name save the oifits files,\n
    `verbose` {bool}:
        If True, print useful informations.

    Returns:
    --------
    `dic` {dict}:
        Oifits formated dictionnary,\n
    `savedfile` {str}:
        Name of the saved oifits file.

    """

    if input_calibrated is None:
        cprint("\nError save : Wrong data format!", on_color="on_red")
        return None

    if oifits_file is None:
        print("Error: oifits filename is not given, please specify oifits_file.")
        return None

    if type(input_calibrated) is not list:
        input_calibrated = [input_calibrated]

    l_dic = []
    for ical in input_calibrated:
        idic = cal2dict(
            ical,
            pa=pa,
            include_vis=include_vis,
            true_flag_v2=true_flag_v2,
            true_flag_t3=true_flag_t3,
            ind_hole=ind_hole,
            snr=snr,
        )
        l_dic.append(idic)
    dic = l_dic[0]

    if not os.path.exists(datadir):
        print("### Create %s directory to save all requested Oifits ###" % datadir)
        os.mkdir(datadir)

    # ------------------------------
    #       Creation OIFITS
    # ------------------------------
    if verbose:
        print("\n\n### Init creation of OI_FITS (%s) :" % (oifits_file))

    # refdate = datetime.datetime(2000, 1, 1)  # Unix time reference
    hdulist = fits.HDUList()

    hdr = dic["info"].get("HDR", {})
    hdu = fits.PrimaryHDU()
    hdu.header["DATE"] = datetime.datetime.now().strftime(
        format="%F"
    )  # , 'Creation date'
    hdu.header["ORIGIN"] = "Sydney University"
    hdu.header["CONTENT"] = "OIFITS2"
    hdu.header["DATE-OBS"] = hdr.get("date-obs", "")
    hdu.header["TELESCOP"] = hdr.get("TELESCOP", "")
    hdu.header["INSTRUME"] = hdr.get("INSTRUME", "")
    hdu.header["OBSERVER"] = hdr.get("OBSERVER", "")
    hdu.header["OBJECT"] = hdr.get("OBJECT", dic["info"]["TARGET"])
    hdu.header["INSMODE"] = "NRM"
    hdu.header["FILT"] = dic["info"]["FILT"]
    hdu.header["MJD"] = dic["info"]["MJD"]
    hdu.header["MASK"] = dic["info"]["MASK"]
    hdu.header["SEEING"] = dic["info"].get("SEEING", "")
    hdu.header["CALIB"] = dic["info"]["CALIB"]

    hdulist.append(hdu)
    # ------------------------------
    #        OI Wavelength
    # ------------------------------

    if verbose:
        print("-> Including OI Wavelength table...")
    data = dic["OI_WAVELENGTH"]

    EFF_WAVE, EFF_BAND = [], []
    for dic in l_dic:
        d = dic["OI_WAVELENGTH"]
        EFF_WAVE.append(d["EFF_WAVE"][0])
        EFF_BAND.append(d["EFF_BAND"][0])

    iwl = len(EFF_WAVE)
    # Data
    # -> Initiation new hdu table :
    hdu = fits.BinTableHDU.from_columns(
        fits.ColDefs(
            (
                fits.Column(
                    name="EFF_WAVE", format="1E", unit="METERS", array=EFF_WAVE
                ),
                fits.Column(
                    name="EFF_BAND", format="1E", unit="METERS", array=EFF_BAND
                ),
            )
        )
    )

    # Header
    hdu.header["EXTNAME"] = "OI_WAVELENGTH"
    hdu.header["OI_REVN"] = 2  # , 'Revision number of the table definition'
    # 'Name of detector, for cross-referencing'
    hdu.header["INSNAME"] = dic["info"]["INSTRUME"]
    hdulist.append(hdu)  # Add current HDU to the final fits file.

    # ------------------------------
    #          OI Target
    # ------------------------------
    if verbose:
        print("-> Including OI Target table...")

    name_star = dic["info"]["TARGET"]

    customSimbad = Simbad()
    customSimbad.add_votable_fields("propermotions", "sptype", "parallax")

    # Add information from Simbad:
    if fake_obj:
        ra = pmra = dec = pmdec = plx = [0]
        spectyp = ["fake"]
    else:
        try:
            query = customSimbad.query_object(name_star)
            coord = SkyCoord(
                query["RA"][0] + " " + query["DEC"][0], unit=(u.hourangle, u.deg)
            )
            ra = [coord.ra.deg]
            dec = [coord.dec.deg]
            spectyp = query["SP_TYPE"]
            pmra = query["PMRA"]
            pmdec = query["PMDEC"]
            plx = query["PLX_VALUE"]
        except Exception:
            ra = pmra = dec = pmdec = plx = [0]
            spectyp = ["fake"]

    hdu = fits.BinTableHDU.from_columns(
        fits.ColDefs(
            (
                fits.Column(name="TARGET_ID", format="1I", array=[1]),
                fits.Column(name="TARGET", format="16A", array=[name_star]),
                fits.Column(name="RAEP0", format="1D", unit="DEGREES", array=ra),
                fits.Column(name="DECEP0", format="1D", unit="DEGREES", array=dec),
                fits.Column(name="EQUINOX", format="1E", unit="YEARS", array=[2000]),
                fits.Column(name="RA_ERR", format="1D", unit="DEGREES", array=[0]),
                fits.Column(name="DEC_ERR", format="1D", unit="DEGREES", array=[0]),
                fits.Column(name="SYSVEL", format="1D", unit="M/S", array=[0]),
                fits.Column(name="VELTYP", format="8A", array=["UNKNOWN"]),
                fits.Column(name="VELDEF", format="8A", array=["OPTICAL"]),
                fits.Column(name="PMRA", format="1D", unit="DEG/YR", array=pmra),
                fits.Column(name="PMDEC", format="1D", unit="DEG/YR", array=pmdec),
                fits.Column(name="PMRA_ERR", format="1D", unit="DEG/YR", array=[0]),
                fits.Column(name="PMDEC_ERR", format="1D", unit="DEG/YR", array=[0]),
                fits.Column(name="PARALLAX", format="1E", unit="DEGREES", array=plx),
                fits.Column(name="PARA_ERR", format="1E", unit="DEGREES", array=[0]),
                fits.Column(name="SPECTYP", format="16A", array=spectyp),
            )
        )
    )

    hdu.header["EXTNAME"] = "OI_TARGET"
    hdu.header["OI_REVN"] = 2, "Revision number of the table definition"
    hdulist.append(hdu)

    # ------------------------------
    #           OI Array
    # ------------------------------

    if verbose:
        print("-> Including OI Array table...")

    staxy = dic["info"]["xycoord"]
    N_ap = len(staxy)
    telName = ["A%i" % x for x in np.arange(N_ap) + 1]
    staName = telName
    diameter = [0] * N_ap

    staxyz = []
    for x in staxy:
        a = list(x)
        line = [a[0], a[1], 0]
        staxyz.append(line)

    staIndex = np.arange(N_ap) + 1

    pscale = dic["info"]["PSCALE"] / 1000.0  # arcsec
    isz = dic["info"]["ISZ"]  # Size of the image to extract NRM data
    fov = [pscale * isz] * N_ap
    fovtype = ["RADIUS"] * N_ap

    hdu = fits.BinTableHDU.from_columns(
        fits.ColDefs(
            (
                fits.Column(name="TEL_NAME", format="16A", array=telName),
                fits.Column(name="STA_NAME", format="16A", array=staName),
                fits.Column(name="STA_INDEX", format="1I", array=staIndex),
                fits.Column(
                    name="DIAMETER", unit="METERS", format="1E", array=diameter
                ),
                fits.Column(name="STAXYZ", unit="METERS", format="3D", array=staxyz),
                fits.Column(name="FOV", unit="ARCSEC", format="1D", array=fov),
                fits.Column(name="FOVTYPE", format="6A", array=fovtype),
            )
        )
    )

    hdu.header["EXTNAME"] = "OI_ARRAY"
    hdu.header["ARRAYX"] = float(0)
    hdu.header["ARRAYY"] = float(0)
    hdu.header["ARRAYZ"] = float(0)
    hdu.header["ARRNAME"] = dic["info"]["MASK"]
    hdu.header["FRAME"] = "SKY"
    hdu.header["OI_REVN"] = 2, "Revision number of the table definition"
    hdulist.append(hdu)

    # ------------------------------
    #           OI VIS
    # ------------------------------
    if include_vis:
        if verbose:
            print("-> Including OI Vis table...")

        data = dic["OI_VIS"]
        if type(data["TARGET_ID"]) is int:
            npts = len(dic["OI_VIS"]["VISAMP"])
        else:
            npts = 1

        staIndex = _format_staindex_v2(data["STA_INDEX"])
        if type(data["MJD"]) is not float:
            mjd = data["MJD"][0]
        else:
            mjd = data["MJD"]

        hdu = fits.BinTableHDU.from_columns(
            fits.ColDefs(
                [
                    fits.Column(
                        name="TARGET_ID", format="1I", array=[data["TARGET_ID"]] * npts
                    ),
                    fits.Column(
                        name="TIME",
                        format="1D",
                        unit="SECONDS",
                        array=[data["TIME"]] * npts,
                    ),
                    fits.Column(
                        name="MJD", unit="DAY", format="1D", array=[data["MJD"]] * npts
                    ),
                    fits.Column(
                        name="INT_TIME",
                        format="1D",
                        unit="SECONDS",
                        array=[data["INT_TIME"]] * npts,
                    ),
                    fits.Column(name="VISAMP", format="1D", array=data["VISAMP"]),
                    fits.Column(name="VISAMPERR", format="1D", array=data["VISAMPERR"]),
                    fits.Column(
                        name="VISPHI",
                        format="1D",
                        unit="DEGREES",
                        array=np.rad2deg(data["VISPHI"]),
                    ),
                    fits.Column(
                        name="VISPHIERR",
                        format="1D",
                        unit="DEGREES",
                        array=np.rad2deg(data["VISPHIERR"]),
                    ),
                    fits.Column(
                        name="UCOORD", format="1D", unit="METERS", array=data["UCOORD"]
                    ),
                    fits.Column(
                        name="VCOORD", format="1D", unit="METERS", array=data["VCOORD"]
                    ),
                    fits.Column(name="STA_INDEX", format="2I", array=staIndex),
                    fits.Column(name="FLAG", format="1L", array=data["FLAG"]),
                ]
            )
        )

        hdu.header["OI_REVN"] = 2, "Revision number of the table definition"
        hdu.header["EXTNAME"] = "OI_VIS"
        hdu.header["INSNAME"] = dic["info"]["INSTRUME"]
        hdu.header["ARRNAME"] = dic["info"]["MASK"]
        hdu.header["DATE-OBS"] = hdr.get("date-obs", "")
        hdulist.append(hdu)

    # ------------------------------
    #           OI VIS2
    # ------------------------------
    if verbose:
        print("-> Including OI Vis2 table...")

    data = dic["OI_VIS2"]
    if type(data["TARGET_ID"]) != np.array:
        npts = len(data["VIS2DATA"])
    else:
        npts = 1

    if type(data["TARGET_ID"]) == int:
        npts = len(dic["OI_VIS2"]["VIS2DATA"])
        targetId = [data["TARGET_ID"]] * npts
        time = [data["TIME"]] * npts
        mjd = [data["MJD"]] * npts
        intTime = [data["INT_TIME"]] * npts
    else:
        npts = 1
        targetId = [1] * len(data["VIS2DATA"])
        time = data["TIME"]
        mjd = data["MJD"]
        intTime = data["INT_TIME"]

    staIndex = _format_staindex_v2(data["STA_INDEX"])

    VIS2DATA, VIS2ERR, FLAG = [], [], []
    for dic in l_dic:
        d = dic["OI_VIS2"]
        VIS2DATA.append(d["VIS2DATA"])
        VIS2ERR.append(d["VIS2ERR"])
        FLAG.append(d["FLAG"])
    VIS2DATA = np.array(VIS2DATA).T
    VIS2ERR = np.array(VIS2ERR).T
    FLAG = np.array(FLAG).T

    hdu = fits.BinTableHDU.from_columns(
        fits.ColDefs(
            [
                fits.Column(name="TARGET_ID", format="1I", array=targetId),
                fits.Column(name="TIME", format="1D", unit="SECONDS", array=time),
                fits.Column(name="MJD", unit="DAY", format="1D", array=mjd),
                fits.Column(
                    name="INT_TIME", format="1D", unit="SECONDS", array=intTime
                ),
                fits.Column(name="VIS2DATA", format="%iD" % iwl, array=VIS2DATA),
                fits.Column(name="VIS2ERR", format="%iD" % iwl, array=VIS2ERR),
                fits.Column(
                    name="UCOORD", format="1D", unit="METERS", array=data["UCOORD"]
                ),
                fits.Column(
                    name="VCOORD", format="1D", unit="METERS", array=data["VCOORD"]
                ),
                fits.Column(name="STA_INDEX", format="2I", array=staIndex),
                fits.Column(name="FLAG", format="%iL" % iwl, array=FLAG),
            ]
        )
    )

    hdu.header["EXTNAME"] = "OI_VIS2"
    hdu.header["INSNAME"] = dic["info"]["INSTRUME"]
    hdu.header["ARRNAME"] = dic["info"]["MASK"]
    hdu.header["OI_REVN"] = 2, "Revision number of the table definition"
    hdu.header["DATE-OBS"] = hdr.get("date-obs", "")
    hdulist.append(hdu)

    # ------------------------------
    #           OI T3
    # ------------------------------
    if verbose:
        print("-> Including OI T3 table...")

    data = dic["OI_T3"]
    try:
        check_oi = type(float(data["MJD"]))
    except TypeError:
        check_oi = int

    if check_oi == float:
        t3phi = dic["OI_T3"]["T3PHI"]
        npts = len(t3phi)
        targetId = np.ones_like(t3phi)
        time = np.zeros_like(t3phi)
        mjd = [data["MJD"]] * npts
        intTime = [data["INT_TIME"]] * npts
    else:
        npts = 1
        targetId = np.ones_like(data["T3PHI"])
        time = data["TIME"]
        mjd = data["MJD"]
        intTime = data["INT_TIME"]

    T3AMP, T3AMPERR, T3PHI, T3PHIERR, FLAG = [], [], [], [], []
    for dic in l_dic:
        d = dic["OI_T3"]
        T3AMP.append(d["T3AMP"])
        T3AMPERR.append(d["T3AMPERR"])
        T3PHI.append(d["T3PHI"])
        T3PHIERR.append(d["T3PHIERR"])
        FLAG.append(d["FLAG"])
    T3AMP = np.array(T3AMP).T
    T3AMPERR = np.array(T3AMPERR).T
    T3PHI = np.array(T3PHI).T
    T3PHIERR = np.array(T3PHIERR).T
    FLAG = np.array(FLAG).T

    staIndex = _format_staindex_t3(data["STA_INDEX"])
    hdu = fits.BinTableHDU.from_columns(
        fits.ColDefs(
            (
                fits.Column(name="TARGET_ID", format="1I", array=targetId),
                fits.Column(name="TIME", format="1D", unit="SECONDS", array=time),
                fits.Column(name="MJD", format="1D", unit="DAY", array=mjd),
                fits.Column(
                    name="INT_TIME", format="1D", unit="SECONDS", array=intTime
                ),
                fits.Column(name="T3AMP", format="%iD" % iwl, array=T3AMP),
                fits.Column(name="T3AMPERR", format="%iD" % iwl, array=T3AMPERR),
                fits.Column(
                    name="T3PHI", format="%iD" % iwl, unit="DEGREES", array=T3PHI
                ),
                fits.Column(
                    name="T3PHIERR", format="%iD" % iwl, unit="DEGREES", array=T3PHIERR
                ),
                fits.Column(
                    name="U1COORD", format="1D", unit="METERS", array=data["U1COORD"]
                ),
                fits.Column(
                    name="V1COORD", format="1D", unit="METERS", array=data["V1COORD"]
                ),
                fits.Column(
                    name="U2COORD", format="1D", unit="METERS", array=data["U2COORD"]
                ),
                fits.Column(
                    name="V2COORD", format="1D", unit="METERS", array=data["V2COORD"]
                ),
                fits.Column(name="STA_INDEX", format="3I", array=staIndex),
                fits.Column(name="FLAG", format="%iL" % iwl, array=FLAG),
            )
        )
    )

    hdu.header["EXTNAME"] = "OI_T3"
    hdu.header["INSNAME"] = dic["info"]["INSTRUME"]
    hdu.header["ARRNAME"] = dic["info"]["MASK"]
    hdu.header["OI_REVN"] = 2, "Revision number of the table definition"
    hdu.header["DATE-OBS"] = hdr.get("date-obs", "")
    hdulist.append(hdu)

    # ------------------------------
    #          Save file
    # ------------------------------
    savedfile = os.path.join(datadir, oifits_file)
    hdulist.writeto(savedfile, overwrite=True)
    if verbose:
        cprint("\n\n### OIFITS CREATED (%s)." % oifits_file, "cyan")

    if len(l_dic) == 1:
        l_dic = l_dic[0]
    return l_dic, savedfile


def _plot_UV(ax1, l_dic, dic_color, diffWl=False):
    l_band_al, l_bmax = [], []
    for dic in l_dic:
        tmp = _apply_flag(dic)
        U = tmp.U
        V = tmp.V
        band = tmp.band
        wl = tmp.wl
        label = fr"{wl * 1e6:2.2f} $\mu m$ ({band})"
        if diffWl:
            c1, c2 = dic_color[band], dic_color[band]
            if band not in l_band_al:
                label = fr"{wl * 1e6:2.2f} $\mu m$ ({band})"
                l_band_al.append(band)
            else:
                label = ""
        else:
            c1, c2 = "#00adb5", "#fc5185"
        l_bmax.append(tmp.bmax)

        ax1.scatter(
            U, V, s=50, c=c1, label=label, edgecolors="#364f6b", marker="o", alpha=1
        )
        ax1.scatter(
            -1 * np.array(U),
            -1 * np.array(V),
            c=c2,
            s=50,
            edgecolors="#364f6b",
            marker="o",
            alpha=1,
        )
    return l_bmax


def _plot_UV_ifu(ax1, fig, l_dic):
    l_bmax = []
    all_U, all_V, all_wl = [], [], []

    for dic in l_dic:
        tmp = _apply_flag(dic)
        U = tmp.U
        V = tmp.V
        wl = tmp.wl
        l_bmax.append(tmp.bmax)
        all_U.append(U)
        all_V.append(V)
        all_wl.append([wl] * len(U))
    all_U = np.array(all_U)
    all_V = np.array(all_V)
    all_wl = np.array(all_wl) * 1e6

    ax1.scatter(
        all_U,
        all_V,
        s=40,
        c=all_wl,
        marker="o",
        edgecolors="#364f6b",
        alpha=1,
        linewidth=0.1,
        cmap="jet",
    )
    sc = ax1.scatter(
        -all_U,
        -all_V,
        s=40,
        c=all_wl,
        marker="o",
        edgecolors="#364f6b",
        alpha=1,
        linewidth=0.1,
        cmap="jet",
    )

    position = fig.add_axes([0.22, 0.95, 0.1, 0.015])
    fig.colorbar(sc, cax=position, orientation="horizontal", drawedges=False)
    ax1.text(
        0.53, 0.98, r"$\lambda$ [µm]", ha="center", va="center", transform=ax1.transAxes
    )
    return l_bmax


def _plot_V2(ax2, l_dic, dic_color, diffWl=False):
    max_f_vis = []
    for dic in l_dic:
        tmp = _apply_flag(dic, unit="arcsec")
        V2 = tmp.vis2
        e_V2 = tmp.e_vis2
        sp_freq_vis = tmp.sp_freq_vis
        max_f_vis.append(np.max(sp_freq_vis))
        band = tmp.band
        if diffWl:
            mfc = dic_color[band]
        else:
            mfc = "#00adb5"

        ax2.errorbar(
            sp_freq_vis,
            V2,
            yerr=e_V2,
            linestyle="None",
            capsize=1,
            mfc=mfc,
            ecolor="#364f6b",
            mec="#364f6b",
            marker=".",
            elinewidth=0.5,
            alpha=1,
            ms=9,
        )
    return max_f_vis


def _plot_V2_ifu(ax2, l_dic):
    max_f_vis = []
    all_V2, all_e_V2, all_freq, all_wl = [], [], [], []
    for dic in l_dic:
        tmp = _apply_flag(dic, unit="arcsec")
        V2 = tmp.vis2
        e_V2 = tmp.e_vis2
        sp_freq_vis = tmp.sp_freq_vis
        max_f_vis.append(np.max(sp_freq_vis))

        all_V2.extend(V2)
        all_e_V2.extend(e_V2)
        all_freq.extend(sp_freq_vis)
        all_wl.extend([tmp.wl] * len(V2))

    all_V2 = np.array(all_V2)
    all_e_V2 = np.array(all_e_V2)
    all_freq = np.array(all_freq)
    all_wl = np.array(all_wl)

    ax2.errorbar(
        all_freq,
        all_V2,
        yerr=all_e_V2,
        linestyle="None",
        capsize=1,
        ecolor="#364f6b",
        mec="#364f6b",
        marker="None",
        elinewidth=0.5,
        alpha=1,
        ms=9,
    )
    ax2.scatter(
        all_freq,
        all_V2,
        s=20,
        c=all_wl,
        zorder=20,
        linewidth=0.5,
        marker="o",
        edgecolors="#364f6b",
        alpha=1,
        cmap="jet",
    )

    return max_f_vis


def _plot_CP(ax3, l_dic, dic_color, conv_cp, diffWl=False):
    max_f_cp = []
    for dic in l_dic:
        tmp = _apply_flag(dic, unit="arcsec")
        cp = tmp.cp * conv_cp
        e_cp = tmp.e_cp * conv_cp
        sp_freq_cp = tmp.sp_freq_cp
        max_f_cp.append(np.max(sp_freq_cp))
        band = tmp.band
        if diffWl:
            mfc = dic_color[band]
        else:
            mfc = "#00adb5"

        ax3.errorbar(
            sp_freq_cp,
            cp,
            yerr=e_cp,
            linestyle="None",
            capsize=1,
            mfc=mfc,
            ecolor="#364f6b",
            mec="#364f6b",
            marker=".",
            elinewidth=0.5,
            alpha=1,
            ms=9,
        )
    return max_f_cp


def _plot_CP_ifu(ax3, l_dic, conv_cp):
    max_f_cp = []
    all_CP, all_e_cp, all_freq, all_wl = [], [], [], []
    for dic in l_dic:
        tmp = _apply_flag(dic, unit="arcsec")
        cp = tmp.cp * conv_cp
        e_cp = tmp.e_cp * conv_cp
        sp_freq_cp = tmp.sp_freq_cp
        max_f_cp.append(np.max(sp_freq_cp))

        all_CP.extend(cp)
        all_e_cp.extend(e_cp)
        all_freq.extend(sp_freq_cp)
        all_wl.extend([tmp.wl] * len(cp))

    ax3.errorbar(
        all_freq,
        all_CP,
        yerr=all_e_cp,
        linestyle="None",
        capsize=1,
        ecolor="#364f6b",
        mec="#364f6b",
        marker="None",
        elinewidth=0.5,
        alpha=1,
        ms=9,
    )
    ax3.scatter(
        all_freq,
        all_CP,
        s=20,
        c=all_wl,
        zorder=20,
        linewidth=0.5,
        marker="o",
        edgecolors="#364f6b",
        alpha=1,
        cmap="jet",
    )
    return max_f_cp


def show(
    inputList,
    diffWl=False,
    ind_hole=None,
    vmin=0,
    vmax=1.05,
    cmax=180,
    setlog=False,
    pa=0,
    unit="arcsec",
    unit_cp="deg",
    snr=4,
    true_flag_v2=True,
    true_flag_t3=False,
):
    """Show oifits data of a multiple dataset (loaded with oifits.load or oifits filename).

    Parameters:
    -----------
    `diffWl` {bool}:
        If True, differentiate the file (wavelenghts) by color,\n
    `ind_hole` {int}:
        By default, ind_hole is None, all the CP are considered ncp = N(N-1)(N-2)/6. If
        ind_hole is set, show only the independant CP including the given hole
        ncp = (N-1)(N-2)/2.\n
    `vmin`, `vmax` {float}:
        Minimum and maximum visibilities (default: 0, 1.05),\n
    `cmax` {float}:
        Maximum closure phase [deg] (default: 180),\n
    `setlog` {bool}:
        If True, the visibility curve is plotted in log scale,\n
    `unit` {str}:
        Unit of the sp. frequencies (default: 'arcsec'),\n
    `unit_cp` {str}:
        Unit of the closure phases (default: 'deg'),\n
    `true_flag_v2` {bool}:
        If inputs are classes from amical.calibrate, compute the true flag of vis2
        using snr parameter (default: True).\n
    `true_flag_t3` {bool}:
        If inputs are classes from amical.calibrate, compute the true flag of cp
        using snr parameter (default: True),\n
    `snr` {float}:
        If inputs are classes from amical.calibrate, use snr param to compute flag,
    """

    if type(inputList) is not list:
        inputList = [inputList]

    try:
        inputList[0].v2
        isclass = True
    except AttributeError:
        isclass = False

    if isclass:
        l_dic = [
            cal2dict(
                x,
                pa=pa,
                true_flag_v2=true_flag_v2,
                ind_hole=ind_hole,
                true_flag_t3=true_flag_t3,
                snr=snr,
            )
            for x in inputList
        ]
        print("\n -- SHOW -- Inputs are classes from amical.calibrate:")
        print("-> (Check true_flag_v2, true_flag_t3 and snr parameters)\n")
    elif type(inputList[0]) is str:
        l_dic = [load(x) for x in inputList]
        print("Inputs are oifits filename.")
    elif type(inputList[0]) is dict:
        l_dic = inputList
        print("Inputs are dict from amical.load.")
    else:
        print("Wrong input format.")
        return None

    dic_color = {}
    i_c = 0
    for dic in l_dic:
        filt = dic["info"]["FILT"]
        if filt not in dic_color.keys():
            dic_color[filt] = list_color[i_c]
            i_c += 1

    fontsize = 14
    fig = plt.figure(figsize=(16, 5.5))
    ax1 = plt.subplot2grid((2, 6), (0, 0), rowspan=2, colspan=2)
    ax2 = plt.subplot2grid((2, 6), (0, 2), colspan=4)
    ax3 = plt.subplot2grid((2, 6), (1, 2), colspan=4)

    # Plot plan UV
    # -------
    ins = l_dic[0]["info"]["INSTRUME"]

    if ("IFS" in ins) & (len(l_dic) > 1):
        l_bmax = _plot_UV_ifu(ax1, fig, l_dic)
    else:
        l_bmax = _plot_UV(ax1, l_dic, dic_color, diffWl=False)

    Bmax = np.max(l_bmax)
    ax1.axis([Bmax, -Bmax, -Bmax, Bmax])
    ax1.spines["left"].set_visible(False)
    ax1.spines["right"].set_visible(False)
    ax1.spines["bottom"].set_visible(False)
    ax1.spines["top"].set_visible(False)
    ax1.patch.set_facecolor("#f7f9fc")
    ax1.patch.set_alpha(1)
    ax1.xaxis.set_ticks_position("none")
    ax1.yaxis.set_ticks_position("none")
    if diffWl:
        handles, labels = ax1.get_legend_handles_labels()
        labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
        ax1.legend(handles, labels, loc="best", fontsize=9)

    unitlabel = {
        "m": "m",
        "rad": "rad$^{-1}$",
        "arcsec": "arcsec$^{-1}$",
        "lambda": r"M$\lambda$",
    }

    ax1.set_xlabel(r"U [%s]" % unitlabel[unit], fontsize=fontsize)
    ax1.set_ylabel(r"V [%s]" % unitlabel[unit], fontsize=fontsize)
    ax1.grid(alpha=0.2)

    # Plot V2
    # -------
    if ("IFS" in ins) & (len(l_dic) > 1):
        max_f_vis = _plot_V2_ifu(ax2, l_dic)
    else:
        max_f_vis = _plot_V2(ax2, l_dic, dic_color, diffWl=diffWl)

    ax2.hlines(1, 0, 1.2 * np.max(max_f_vis), lw=1, color="k", alpha=0.2, ls="--")

    ax2.set_ylim([vmin, vmax])
    ax2.set_xlim([0, 1.2 * np.max(max_f_vis)])
    ax2.set_ylabel(r"$V^2$", fontsize=fontsize)
    ax2.spines["left"].set_visible(False)
    ax2.spines["right"].set_visible(False)
    ax2.spines["bottom"].set_visible(False)
    ax2.spines["top"].set_visible(False)
    ax2.patch.set_facecolor("#f7f9fc")
    ax2.patch.set_alpha(1)
    ax2.xaxis.set_ticks_position("none")
    ax2.yaxis.set_ticks_position("none")

    if setlog:
        ax2.set_yscale("log")
    ax2.grid(which="both", alpha=0.2)

    # Plot CP
    # -------
    if unit_cp == "rad":
        conv_cp = np.pi / 180.0
        h1 = np.pi
    else:
        conv_cp = 1
        h1 = np.rad2deg(np.pi)

    cmin = -cmax

    if ("IFS" in ins) & (len(l_dic) > 1):
        max_f_cp = _plot_CP_ifu(ax3, l_dic, conv_cp)
    else:
        max_f_cp = _plot_CP(ax3, l_dic, dic_color, conv_cp, diffWl=diffWl)

    ax3.hlines(h1, 0, 1.2 * np.max(max_f_cp), lw=1, color="k", alpha=0.2, ls="--")
    ax3.hlines(-h1, 0, 1.2 * np.max(max_f_cp), lw=1, color="k", alpha=0.2, ls="--")
    ax3.spines["left"].set_visible(False)
    ax3.spines["right"].set_visible(False)
    ax3.spines["bottom"].set_visible(False)
    ax3.spines["top"].set_visible(False)
    ax3.patch.set_facecolor("#f7f9fc")
    ax3.patch.set_alpha(1)
    ax3.xaxis.set_ticks_position("none")
    ax3.yaxis.set_ticks_position("none")
    ax3.set_xlabel("Spatial frequency [arcsec$^{-1}$]", fontsize=fontsize)
    ax3.set_ylabel(r"Clos. $\phi$ [%s]" % unit_cp, fontsize=fontsize)
    ax3.axis([0, 1.2 * np.max(max_f_cp), cmin * conv_cp, cmax * conv_cp])
    ax3.grid(which="both", alpha=0.2)

    plt.subplots_adjust(
        top=0.974, bottom=0.1, left=0.05, right=0.99, hspace=0.127, wspace=0.35
    )

    return fig



import pickle
import time
from multiprocessing import Pool

import corner
import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as interp
from scipy.optimize import leastsq
from tqdm import tqdm

"""------------------------------------------------------------------------
cp_tools.py - a collection of functions useful for closure phase analysis
in Python. This includes mas2rad, rad2mas and phase_binary from pysco;
it depends on PyMultiNest, MultiNest and emcee
------------------------------------------------------------------------"""


def mas2rad(x):
    """Convenient little function to convert milliarcsec to radians"""
    return x * np.pi / (180 * 3600 * 1000)


# =========================================================================
# =========================================================================


def rad2mas(x):
    """Convenient little function to convert radians to milliarcseconds"""
    return x / np.pi * (180 * 3600 * 1000)


# =========================================================================
# =========================================================================


def phase_binary(u, v, wavel, p, return_cvis=False):
    """Calculate the phases observed by an array on a binary star
    ----------------------------------------------------------------
    p: 3-component vector (+2 optional), the binary "parameters":
    - p[0] = sep (mas)
    - p[1] = PA (deg) E of N.
    - p[2] = contrast ratio (primary/secondary)

    optional:
    - p[2:] = contrast ratio for several wavelengths that we want
            to calculate the cps over

    - u,v: baseline coordinates (meters)
    - wavel: wavelength (meters)
    ----------------------------------------------------------------"""
    p = np.array(p)
    # relative locations
    th = (p[1] + 90.0) * np.pi / 180.0
    ddec = mas2rad(p[0] * np.sin(th))
    dra = -mas2rad(p[0] * np.cos(th))

    # decompose into two "luminosities"
    # but first, a little trick so this works whether
    # p is a single value or a list of contrasts
    spec = p[2:]
    if len(spec) == 1:
        spec = spec[0]
    l2 = 1.0 / (spec + 1)
    l1 = 1 - l2

    # phase-factor
    output_shape = list(u.shape)
    output_shape[-1] = np.size(wavel)
    phi = np.zeros(output_shape, dtype=complex)
    phi.real = np.cos(-2 * np.pi * (u * dra + v * ddec) / wavel)
    phi.imag = np.sin(-2 * np.pi * (u * dra + v * ddec) / wavel)

    cvis = l1 + l2 * phi

    phase = np.angle(cvis, deg=True)
    if return_cvis:
        return cvis
    else:
        return np.mod(phase + 10980.0, 360.0) - 180.0


# =========================================================================
# =========================================================================


def cp_loglikelihood(params, u, v, wavel, t3data, t3err, model="constant"):
    """Calculate loglikelihood for closure phase data.
    Used both in the MultiNest and MCMC Hammer implementations."""

    # hacky way to introduce priors
    if (params[2] > 5000) or (params[2] < 0.0):
        return -np.inf
    # if (params[0] > 250.) or (params[0] < 0.):
    #    return -np.inf
    if (params[1] > 380.0) or (params[1] < -5.0):
        return -np.inf

    cps = cp_model(params, u, v, wavel, model=model)
    chi2 = np.sum(((t3data - cps) / t3err) ** 2)
    loglike = -chi2 / 2
    return loglike


# =========================================================================
# =========================================================================


def cp_loglikelihood_cov(params, u, v, wavel, t3data, cov_inv, model="constant"):
    """Calculate loglikelihood for closure phase data. Uses the inverse
    covariance matrix rather than the uncertainties
    Used both in the MultiNest and MCMC Hammer implementations."""

    # hacky way to introduce priors
    if (params[2] > 5000) or (params[2] < 0.0):
        return -np.inf
    if (params[0] > 250.0) or (params[0] < 0.0):
        return -np.inf
    if (params[1] > 380.0) or (params[1] < -5):
        return -np.inf

    cps = cp_model(params, u, v, wavel, model=model)
    resids = t3data - cps
    # Loop through wavelengths and calculate the chi2 for each one and add them
    chi2 = 0
    for wav in range(wavel.size):
        # We want obs.T * Cov**-1 * obs
        # But since obs is not a vector, we would need to loop over the second dimension.
        # Instead, here's a trick to do this in a faster way
        temp = resids[:, :, wav].transpose().dot(cov_inv[:, :, wav])
        chi2 += np.sum(temp.transpose() * resids[:, :, wav])
    loglike = -chi2 / 2
    return loglike


# =========================================================================
# =========================================================================


def cp_loglikelihood_proj(
    params, u, v, wavel, proj_t3data, proj_t3err, proj, model="constant"
):
    """Calculate loglikelihood for projected closure phase data.
    Used both in the MultiNest and MCMC Hammer implementations.
    Here proj is the eigenvector array"""

    # hacky way to introduce priors
    #    if (params[2] > 50000) or (params[2] < 0.):
    #        return -np.inf
    if (params[0] > 350.0) or (params[0] < 0.0):
        return -np.inf
    if (params[1] > 380.0) or (params[1] < -5.0):
        return -np.inf

    cps = cp_model(params, u, v, wavel, model=model)

    proj_mod_cps = project_cps(cps, proj)

    chi2 = np.sum(((proj_t3data - proj_mod_cps) / proj_t3err) ** 2)

    loglike = -chi2 / 2
    return loglike


# =========================================================================
# =========================================================================


def cp_loglikelihood_multiple(
    params, u, v, wavel, t3data, t3err, model="constant", ncomp=1
):
    """Calculate loglikelihood for closure phase data and multiple companions.
    Used both in the MultiNest and MCMC Hammer implementations."""
    cps = cp_model(params[0:3], u, v, wavel, model=model)
    for ix in range(1, ncomp):
        # 3 since there are 3 parameters per companion
        cps += cp_model(params[ix * 3 : (ix + 1) * 3], u, v, wavel, model=model)

    chi2 = np.sum(((t3data.ravel() - cps.ravel()) / t3err.ravel()) ** 2)
    loglike = -chi2 / 2
    return loglike


# =========================================================================
# =========================================================================


def cp_model(params, u, v, wavels, model="constant"):
    """Function to model closure phases. Takes a parameter list, u,v triangles and range of wavelengths.
    Allows fitting of a model to contrast vs wavelength.
    Models for contrast ratio:
            constant (contrast is constant with wavelength, default)
            linear (params[2,3]=contrast ratios at end wavelengths),
            free (params[2:]=contrast ratios).
            ndof (the wavelength channels are evenly spaced cubic interpolations in params[2:])
            polynomial (of the form Sum[n] params[n+2]*(wavelength*1e6)**n )
    NOTE: This doesn't allow for nonzero size of each component!"""
    nwav = wavels.size
    model_params = np.zeros(nwav + 2)
    model_params[0:2] = params[0:2]
    if model == "constant":
        cons = np.repeat(params[2], nwav)
    elif model == "linear":
        cons = params[2] + (params[3] - params[2]) * (wavels - wavels[0]) / (
            wavels[-1] - wavels[0]
        )
    elif model == "ndof":
        ndof = params[2:].size
        wavs = np.linspace(np.min(wavels), np.max(wavels), ndof)
        f = interp.interp1d(wavs, params[2:], kind="linear")
        cons = f(wavels)
    elif model == "free":
        # no model, crat vs wav is free to vary.
        cons = params[2:]
    elif model == "polynomial":
        coefficients = params[2:]
        ndof = len(coefficients)
        cons = np.repeat(0.0, nwav)
        xax = (wavels - np.min(wavels)) / (np.max(wavels) - np.min(wavels))
        for order in range(ndof):
            cons += coefficients[order] * xax ** order
    else:
        raise NameError("Unknown model input to cp_model")
    model_params[2:] = cons
    # vectorize the arrays to speed up multi-wavelength calculations
    u = u[..., np.newaxis]  # (ncp x n_runs x 3 x 1) or (ncp x 3 x 1)
    v = v[..., np.newaxis]  # (ncp x n_runs x 3 x 1) or (ncp x 3 x 1)
    # (1 x 1 x 1 x nwav) or (1x1xnwav)
    wavels = wavels[np.newaxis, np.newaxis, :]
    if u.ndim == 4:
        wavels = wavels[np.newaxis]
    phases = phase_binary(u, v, wavels, model_params)
    cps = np.sum(phases, axis=-2)

    return cps


# =========================================================================
# =========================================================================


def project_cps(cps, proj):
    """Short wrapper program to do the projection of a set of closure phases
    onto another basis. proj is the projection matrix, usually the eigenvectors
    for projecting onto a statistically independent basis set."""
    proj_cps = np.zeros((proj.shape[2], cps.shape[1], cps.shape[2]))
    for wav in range(cps.shape[2]):
        proj_cps[:, :, wav] = np.dot(proj[wav].transpose(), cps[:, :, wav])

    return proj_cps


# =========================================================================
# =========================================================================


def hammer(
    cpo,
    ivar=[52.0, 192.0, 1.53],
    ndim="Default",
    nwalcps=50,
    plot=False,
    projected=False,
    niters=1000,
    threads=1,
    model="constant",
    sep_prior=None,
    pa_prior=None,
    crat_prior=None,
    err_scale=1.0,
    extra_error=0.0,
    use_cov=False,
    burn_in=0,
    verbose=False,
):
    import emcee

    """Default implementation of emcee, the MCMC Hammer, for closure phase
    fitting. Requires a closure phase object cpo, and is best called with
    ivar chosen to be near the peak - it can fail to converge otherwise.
    Also allows fitting of a contrast vs wavlength model. See cp_model for details!
    Prior ranges introduce a flat (tophat) prior between the two values specified
    burn_in = the number of iterations to discard due to burn-in"""
    if ndim == "Default":
        ndim = len(ivar)

    ivar = np.array(ivar)  # initial parameters for model-fit

    # Starting parameters for the walkers
    p0 = []
    scatter = np.zeros(ndim) + 0.01
    scatter[0] = 0.05
    scatter[1] = 0.05
    for walker_ix in range(nwalcps):
        p0.append(ivar + ivar * scatter * np.random.rand(ndim))
    #    p0 = [ivar + 0.1*ivar*np.random.rand(ndim) for i in range(nwalcps)] # initialise walcps in a ball
    #    p0 = [ivar + 0.75*ivar*np.random.rand(ndim) for i in range(nwalcps)] # initialise walcps in a ball
    # print('\n -- Running emcee --')

    t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
    t3err *= err_scale

    t0 = time.time()
    if projected is False:
        sampler = emcee.EnsembleSampler(
            nwalcps,
            ndim,
            cp_loglikelihood,
            args=[cpo.u, cpo.v, cpo.wavel, cpo.t3data, t3err, model],
            threads=threads,
        )
    elif use_cov:
        sampler = emcee.EnsembleSampler(
            nwalcps,
            ndim,
            cp_loglikelihood_cov,
            args=[cpo.u, cpo.v, cpo.wavel, cpo.t3data, cpo.cov_inv, model],
            threads=threads,
        )
    else:
        proj_t3err = np.sqrt(cpo.proj_t3err ** 2 + extra_error ** 2)
        proj_t3err *= err_scale
        sampler = emcee.EnsembleSampler(
            nwalcps,
            ndim,
            cp_loglikelihood_proj,
            args=[
                cpo.u,
                cpo.v,
                cpo.wavel,
                cpo.proj_t3data,
                proj_t3err,
                cpo.proj,
                model,
            ],
            threads=threads,
        )

    sampler.run_mcmc(p0, niters, progress=True)
    tf = time.time()

    if verbose:
        print("Time elapsed =", tf - t0, "s")

    chain = sampler.flatchain

    # Remove the burn in
    chain = chain[burn_in:]

    seps = chain[:, 0]
    ths = chain[:, 1]
    cs = chain[:, 2:][:, 0]

    # Now introduce the prior, by ignoring values outside of the range
    if sep_prior is not None:
        wh = (seps >= sep_prior[0]) & (seps <= sep_prior[1])
        seps = seps[wh]
        ths = ths[wh]
        cs = cs[wh]
    if pa_prior is not None:
        wh = (ths >= pa_prior[0]) & (ths <= pa_prior[1])
        seps = seps[wh]
        ths = ths[wh]
        cs = cs[wh]
    if crat_prior is not None:
        wh = (cs >= crat_prior[0]) & (cs <= crat_prior[1])
        seps = seps[wh]
        ths = ths[wh]
        cs = cs[wh]
    # if crat_prior is not None:
    #     #     for ix in range(ndim-2):
    #     #         c = cs[:, ix]
    #     wh = (cs[:, 0] >= crat_prior[0]) & (cs[:, 0] <= crat_prior[1])
    #     seps = seps[wh]
    #     ths = ths[wh]
    #     cs = cs[wh, :]

    # check that there are still some points left!
    if seps.size > 0:
        ngood = len(seps)
        chain = np.zeros((ngood, ndim))
        chain[:, 0] = seps
        chain[:, 1] = ths
        chain[:, 2] = cs
    else:
        print("WARNING: Your priors eliminated all points!")

    meansep = np.mean(seps)
    dsep = np.std(seps)

    meanth = np.mean(ths)
    dth = np.std(ths)

    meanc = np.mean(cs, axis=0)
    dc = np.std(cs, axis=0)

    if verbose:
        print("Separation", meansep, "pm", dsep, "mas")
        print("Position angle", meanth, "pm", dth, "deg")
        print("Contrast", meanc[0], "pm", dc[0])

    if model == "linear":
        print("Contrast2", meanc[1], "pm", dc[1])
        extra_pars = ["Contrast "]
        extra_dims = ["Ratio"]
    elif model == "free":
        extra_pars = np.repeat("Contrast", ndim - 2)
        extra_dims = np.repeat("Ratio", ndim - 2)
    else:
        extra_pars = "None"
        extra_dims = "None"

        paramdims = ["(mas)", "(deg)", "Ratio"]
        for ix, par in enumerate(extra_pars):
            # paramnames.append(par)
            paramdims.append(extra_dims[ix])

        res_p, err_p, err_m = [], [], []
        for i in range(ndim):
            mcmc = np.percentile(chain[:, i], [16, 50, 84])
            q = np.diff(mcmc)
            res_p.append(mcmc[1])
            err_m.append(q[0])
            err_p.append(q[1])

    if plot:
        # plt.figure(figsize=(5, 7))
        # plt.subplot(3, 1, 1)
        # plt.plot(sampler.chain[:, :, 0].T, color='grey', alpha=.5)
        # # plt.plot(len(chain_sep), sep, marker='*', color='#0085ca', zorder=1e3)
        # plt.ylabel('Separation [mas]')
        # plt.subplot(3, 1, 2)
        # plt.plot(sampler.chain[:, :, 1].T, color='grey', alpha=.5)
        # # plt.plot(len(chain_sep), pa, marker='*', color='#0085ca', zorder=1e3)
        # plt.ylabel('PA [deg]')
        # plt.subplot(3, 1, 3)
        # plt.plot(sampler.chain[:, :, 2].T, color='grey', alpha=.2)
        # # plt.plot(len(chain_sep), cr, marker='*', color='#0085ca', zorder=1e3)
        # plt.xlabel('Step')
        # plt.ylabel('CR')
        # plt.tight_layout()
        # plt.show(block=False)

        fig = corner.corner(
            chain,
            labels=["SEP [mas]", "PA [deg]", "CONTRAST"],
            quantiles=(0.16, 0.84),
            show_titles=True,
            title_kwargs={"fontsize": 10},
            color="#096899",
        )
        axes = np.array(fig.axes).reshape((ndim, ndim))
        # Loop over the diagonal
        for i in range(ndim):
            ax = axes[i, i]
            # ax.axvline(value1[i], color="g")
            ax.axvline(res_p[i], color="#ce0056e6", lw=1)

        # Loop over the histograms
        for yi in range(ndim):
            for xi in range(yi):
                ax = axes[yi, xi]
                ax.axvline(res_p[xi], color="#ce0056e6", lw=1)
                ax.axhline(res_p[yi], color="#ce0056e6", lw=1)
                ax.plot(res_p[xi], res_p[yi], "#ce0056e6", lw=1)
        plt.tight_layout()
        plt.show(block=False)

    data = {
        "sep": meansep,
        "delsep": dsep,
        "pa": meanth,
        "delpa": dth,
        "con": meanc,
        "delcon": dc,
        "chain": sampler.chain,
    }
    data2 = {
        "sep": res_p[0],
        "delsepm": err_m[0],
        "delsepp": err_p[0],
        "pa": res_p[1],
        "delpam": err_m[1],
        "delpap": err_p[1],
        "cr": res_p[2],
        "delcrm": err_m[2],
        "delcrp": err_p[2],
        "chain": sampler.chain,
    }

    # and clean up
    # if threads > 1:
    #     sampler.pool.terminate()
    return data, data2


# =========================================================================
# =========================================================================


# def nest(cpo, paramlimits=[20., 250., 0., 360., 1.0001, 10], resume=False, eff=0.3, multi=True,
#          err_scale=1., extra_error=0., plot=True, npoints=1000, logcrat_prior=True,
#          use_cov=False, projected=False, model='constant'):
#     '''Default implementation of a MultiNest fitting routine for closure
#     phase data. Requires a closure phase cpo object, parameter limits and
#     sensible keyword arguments for the multinest parameters.

#     This function does very naughty things creating functions inside this
#     function because PyMultiNest is very picky about how you pass it
#     data.

#     Optional parameter eff tunes sampling efficiency, and multi toggles multimodal
#     nested sampling on and off. Turning off multimodal sampling results in a speed
#     boost of ~ 20-30%.

#     logcrat_prior allows the prior for the contrast ratios to be flat in logarithmic space

#     paramlimits is needed here, and must be a list with length = ncomp,
#     with paramlimits[companion_number]=[minsep,maxsep,minpa,maxpa,mincrat,maxcrat]
#     '''
#     import pymultinest

#     if projected:
#         proj_t3err = np.sqrt(cpo.proj_t3err**2 + extra_error**2)
#         proj_t3err *= err_scale
#     else:
#         t3err = np.sqrt(cpo.t3err**2+extra_error**2)
#         t3err *= err_scale

#     if logcrat_prior:
#         def myprior(cube, ndim, n_params, paramlimits=paramlimits):
#             cube[0] = cube[0]*(paramlimits[1] - paramlimits[0])+paramlimits[0]
#             cube[1] = cube[1]*(paramlimits[3] - paramlimits[2])+paramlimits[2]
#             for ix in range(n_params-2):
#                 cube[ix+2] = 10**(cube[ix+2]*(np.log10(paramlimits[5]) -
#                                               np.log10(paramlimits[4]))+np.log10(paramlimits[4]))
#     else:
#         def myprior(cube, ndim, n_params, paramlimits=paramlimits):
#             cube[0] = cube[0]*(paramlimits[1] - paramlimits[0])+paramlimits[0]
#             cube[1] = cube[1]*(paramlimits[3] - paramlimits[2])+paramlimits[2]
#             for ix in range(n_params-2):
#                 cube[ix+2] = cube[ix+2] * \
#                     (paramlimits[5] - paramlimits[4])+paramlimits[4]

#     if projected:
#         def myloglike(cube, ndim, n_params):
#             loglike = cp_loglikelihood_proj(cube[0:n_params], cpo.u, cpo.v, cpo.wavel,
#                                             cpo.proj_t3data, proj_t3err, cpo.proj, model=model)
#             return loglike

#     elif use_cov:
#         def myloglike(cube, ndim, n_params):
#             loglike = cp_loglikelihood_cov(cube[0:n_params], cpo.u, cpo.v, cpo.wavel,
#                                            cpo.t3data, cpo.cov_inv, model=model)
#             return loglike

#     else:
#         def myloglike(cube, ndim, n_params):
#             loglike = cp_loglikelihood(
#                 cube[0:n_params], cpo.u, cpo.v, cpo.wavel, cpo.t3data, t3err, model=model)
#             return loglike

#     # How many parameters?
#     if model == 'constant':
#         parameters = ['Separation', 'Position Angle', 'Contrast Ratio']
#     elif model == 'free':
#         parameters = ['Separation', 'Position Angle']
#         parameters.extend(len(cpo.wavel)*['Contrast Ratio'])
#     else:
#         raise Exception('Model not yet implemented in nest!')

#     n_params = len(parameters)
#     ndim = n_params

#     # Check that the "chains" directory exists (which multinest needs)
#     if os.path.exists(os.getcwd()+'/chains/') is False:
#         os.mkdir(os.getcwd()+'/chains/')

#     tic = time.time()  # start timing

#     # ---------------------------------
#     # now run MultiNest!
#     # ---------------------------------

#     pymultinest.run(myloglike, myprior, n_params, wrapped_params=[1], resume=resume, verbose=True,
#                     sampling_efficiency=eff, multimodal=multi, n_iter_before_update=1000,
#                     n_live_points=npoints)
#     # let's analyse the results
#     a = pymultinest.Analyzer(n_params=n_params)
#     s = a.get_stats()

#     toc = time.time()

#     if toc-tic < 60.:
#         print('Time elapsed =', toc-tic, 's')
#     else:
#         print('Time elapsed =', (toc-tic)/60., 'mins')

#     print()
#     print("-" * 30, 'ANALYSIS', "-" * 30)
#     print("Global Evidence:\n\t%.15e +- %.15e" %
#           (s['global evidence'], s['global evidence error']))
#     print('')

#     params = s['marginals']
#     print_line = "{0}: {1:.3F} pm {2:.3F}"

#     for param_ix in range(n_params):
#         print(print_line.format(parameters[param_ix], params[param_ix]['median'],
#                                 params[param_ix]['sigma']))
#     if plot:
#         p = pymultinest.PlotMarginalModes(a)
#         plt.figure(figsize=(5*n_params, 5*n_params))
#         for i in range(n_params):
#             plt.subplot(n_params, n_params, n_params * i + i + 1)
#             p.plot_marginal(i, with_ellipses=True,
#                             with_points=False, grid_points=20)
#             plt.ylabel("Probability")
#             plt.xlabel(parameters[i])

#             for j in range(i):
#                 plt.subplot(n_params, n_params, n_params * j + i + 1)
#                 # plt.subplots_adjust(left=0, bottom=0, right=0, top=0, wspace=0, hspace=0)
#                 p.plot_conditional(i, j, with_ellipses=False, with_points=True, grid_points=20)
#                 plt.xlabel(parameters[i])
#                 plt.ylabel(parameters[j])

#     return [a, s]

# =========================================================================
# =========================================================================


def detec_sim_loopfit(everything):
    """Function for multiprocessing in detec_limits. Takes a
    single separation and full angle, contrast lists.
    For each sep,pa,contrast, it calculates 10,000 simulations of that binary
     (adding noise to each). A detection is defined as having chi2_bin - chi2_null <0
     It then counts the number of detections over all separations and"""
    detec_count = np.zeros((everything["nth"], everything["ncon"]))
    ndim = len(everything["error"].shape)

    # This can be done once since it doesn't change with the binary params
    # error should be ncp x nwav, rands should be ncp x nwav x n
    err = everything["error"]
    resids = err[..., np.newaxis] * everything["rands"]

    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):

            bin_cp = cp_model(
                [everything["sep"], th, con],
                everything["u"],
                everything["v"],
                everything["wavel"],
            )
            rnd_cp = bin_cp[..., np.newaxis] + resids

            # We want the difference in chi2 between the binary and null hypothesis.
            #  i.e. using rnd_cp for the single star and rnd_cp-bin_cp for the binary
            #  but this simplifies to the following equation
            chi2_diff = np.sum(
                (resids ** 2 - rnd_cp ** 2) / everything["error"][..., np.newaxis] ** 2,
                axis=tuple(range(ndim)),
            )

            #            chi2_sngl = np.sum(np.sum((((rnd_cp)/ everything['error'][:,:,np.newaxis])**2),axis=0),axis=0)
            #            chi2_binr = np.sum(np.sum((((rnd_cp-bin_cp[:,:,np.newaxis]) / everything['error'][:,:,np.newaxis])**2),axis=0),axis=0)
            #            chi2_diff=chi2_binr-chi2_sngl
            # this counts the number of detections
            detec_count[j, k] = (chi2_diff < (-0.0)).sum()

    # print('done one separation')
    # print(err.shape,bin_cp.shape,rnd_cp.shape,everything['rands'].shape)
    return detec_count


# =========================================================================
# =========================================================================


def detec_sim_loopfit_cov(everything):
    """Function for multiprocessing in detec_limits. Takes a
    single separation and full angle, contrast lists.
    For each sep,pa,contrast, it calculates 10,000 simulations of that binary
     (adding noise to each). A detection is defined as having chi2_bin - chi2_null <0
     It then counts the number of detections over all separations and"""
    detec_count = np.zeros((everything["nth"], everything["ncon"]))
    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):
            bin_cp = cp_model(
                [everything["sep"], th, con],
                everything["u"],
                everything["v"],
                everything["wavel"],
            )
            errs = everything["rands"]
            # binary cp model
            # ----------------------
            rnd_cp = bin_cp[:, :, np.newaxis] + errs
            chi2_sngl = rnd_cp.transpose().dot(everything["cov_inv"]).dot(rnd_cp)
            chi2_binr = errs.transpose().dot(everything["cov_inv"]).dot(errs)
            chi2_diff = chi2_binr - chi2_sngl
            # this counts the number of detections
            detec_count[j, k] = (chi2_diff < (-0.0)).sum()

    # print('done one separation')
    return detec_count


# =========================================================================
# =========================================================================


def detec_sim_loopfit_proj(everything):
    """Function for multiprocessing in detec_limits. Takes a
    single separation and full angle, contrast lists. Made for projected data"""
    detec_count = np.zeros((everything["nth"], everything["ncon"]))
    proj = everything["proj"]
    ndim = len(everything["error"].shape)

    # This can be done once since it doesn't change with the binary params
    # Note that error and rands are already in the projected basis
    proj_resids = everything["error"][..., np.newaxis] * everything["rands"]

    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):

            bin_cp = cp_model(
                [everything["sep"], th, con],
                everything["u"],
                everything["v"],
                everything["wavel"],
            )

            # Project the data:
            proj_bin_cp = project_cps(bin_cp, proj)
            proj_rnd_cp = proj_bin_cp[..., np.newaxis] + proj_resids

            # We want the difference in chi2 between the binary and null hypothesis.
            #  i.e. using rnd_cp for the single star and rnd_cp-bin_cp for the binary
            #  but this simplifies to the following equation
            chi2_diff = np.sum(
                (proj_resids ** 2 - proj_rnd_cp ** 2)
                / everything["error"][..., np.newaxis] ** 2,
                axis=tuple(range(ndim)),
            )

            #            chi2_sngl = np.sum(np.sum((((rnd_cp)/ everything['error'][:,:,np.newaxis])**2),axis=0),axis=0)
            #            chi2_binr = np.sum(np.sum((((rnd_cp-bin_cp[:,:,np.newaxis]) / everything['error'][:,:,np.newaxis])**2),axis=0),axis=0)
            #            chi2_diff=chi2_binr-chi2_sngl
            # this counts the number of detections
            detec_count[j, k] = (chi2_diff < (-0.0)).sum()

    return detec_count


# =========================================================================
# =========================================================================


def detec_limits(
    cpo,
    nsim=2000,
    nsep=32,
    nth=20,
    ncon=32,
    smin="Default",
    smax="Default",
    cmin=1.0001,
    cmax=500.0,
    extra_error=0,
    threads=0,
    save=False,
    projected=False,
    use_cov=False,
    icpo=False,
    err_scale=1.0,
    no_plot=False,
    linear_in_mags=False,
):
    """uses a Monte Carlo simulation to establish contrast-separation
    detection limits given an array of standard deviations per closure phase.

    Because different separation-contrast grid points are entirely
    separate, this task is embarrassingly parallel. If you want to
    speed up the calculation, use multiprocessing with a threads
    argument equal to the number of available cores.

    Make nseps a multiple of threads! This uses the cores most efficiently.

    Hyperthreading (2x processes per core) in my experience gets a ~20%
    improvement in speed.

    Written by F. Martinache and B. Pope.
    ACC added option for projected data and a few tweaks.

    Use_cov option allows the random clps to be generated using the sample covariance matrix.

    Note also that the calculation of the model closure phases could be done outside
    the big loop, which would be efficient on CPU but not RAM. However, ACC
    tried adding this and ran out of RAM (8GB) on GPI data (8880 clps), so removed it."""

    # Note that the accuracy of these sims are limited by the number of fake clps sets you take.
    # e.g. if you only have 10 sims you can't get accuracy better than 10%
    # (and even then, you will need several times more than this to get a robust 10% limit).
    print("Detection limit resolution:", 100.0 / (nsim * nth), "%")
    if 100.0 / (nsim * nth) > 0.01:
        print(
            "It is recommended that you increase nsim if you want robust 99.9% detection limits."
        )

    # ------------------------
    # first, load your data!
    # ------------------------
    if projected is True:
        proj = cpo.proj
        error = np.sqrt(cpo.proj_t3err ** 2 + extra_error ** 2) * err_scale
        n_clps = cpo.proj.shape[-1]
        n_runs = cpo.n_runs
    elif icpo is True:
        proj = []
        cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
        error = cpo.t3err * err_scale
        n_clps = cpo.n_clps
        n_runs = cpo.n_runs
    else:
        proj = []
        cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
        error = cpo.t3err * err_scale
        n_clps = cpo.ndata
        n_runs = 1

    if use_cov:
        cov_inv = cpo.cov_inv
    else:
        cov_inv = []

    # nwav = cpo.wavel.size
    # ndata = cpo.u.shape[0]
    # u=np.repeat(np.resize(cpo.u,[ndata,3,1]),nwav,2).ravel()
    # v=np.repeat(np.resize(cpo.v,[ndata,3,1]),nwav,2).ravel()
    # wavel=np.repeat(np.repeat(np.resize(cpo.wavel,[1,1,nwav]),ndata,0),3,1).ravel()
    wavel = cpo.wavel
    u = cpo.u
    v = cpo.v

    w = np.array(np.sqrt(u ** 2 + v ** 2)) / np.median(wavel)

    if smin == "Default":
        smin = rad2mas(1.0 / 4 / np.max(w))

    if smax == "Default":
        smax = rad2mas(1.0 / np.min(w))

    # ------------------------
    # initialise Monte Carlo
    # ------------------------

    seps = smin + (smax - smin) * np.linspace(0, 1, nsep)
    ths = 0.0 + 360.0 * np.linspace(0, 1, nth)
    cons = cmin + (cmax - cmin) * np.linspace(0, 1, ncon)
    if linear_in_mags:
        cons = 10 ** (
            np.linspace(2.5 * np.log10(cmin), 2.5 * np.log10(cmax), ncon) / 2.5
        )

    # Generate the random numbers first to save time
    rands_shape = list(error.shape)
    rands_shape.append(nsim)
    if projected:
        # The random numbers should be in the projected basis
        rands = np.random.normal(size=rands_shape)

    elif use_cov:
        # use full multivariate gaussian with sample covariance to get rands. Dont use with projected!
        rands = np.zeros((n_clps * n_runs, nsim))
        for run in range(n_runs):
            rands[n_clps * (run) : n_clps * (run + 1), :] = np.transpose(
                np.random.multivariate_normal(
                    np.zeros(n_clps), np.atleast_3d(cpo.sample_cov)[:, :, run], (nsim)
                )
            )
    else:
        rands = np.random.normal(size=rands_shape)
    print(rands.shape, error.shape)
    all_vars = []
    print("Setting up the big dictionary to store variables for loop")
    for ix in range(nsep):
        everything = {
            "sep": seps[ix],
            "cons": cons,
            "ths": ths,
            "ix": ix,
            "nsep": nsep,
            "ncon": ncon,
            "nth": nth,
            "u": cpo.u,
            "v": cpo.v,
            "nsim": nsim,
            "rands": rands,
            "error": error,
            "wavel": cpo.wavel,
            "proj": proj,
            "n_clps": n_clps,
            "n_runs": n_runs,
            "cov_inv": cov_inv,
        }
        all_vars.append(everything)

    print("Starting big loop over separations")
    # ------------------------
    # Run Monte Carlo
    # ------------------------
    tic = time.time()  # start the clock'
    if threads == 0:
        ndetec = np.zeros((nsep, nth, ncon))
        for ix, sep in enumerate(seps):
            if projected is True:
                ndetec[ix, :, :] = detec_sim_loopfit_proj(all_vars[ix])
            elif use_cov:
                ndetec[ix, :, :] = detec_sim_loopfit_cov(all_vars[ix])
            else:
                ndetec[ix, :, :] = detec_sim_loopfit(all_vars[ix])
            toc = time.time()
            if ix != 10:
                remaining = (toc - tic) * (nsep - ix) / float(ix + 1)
                if remaining > 60:
                    print("Estimated time remaining: %.2f mins" % (remaining / 60.0))
                else:
                    print("Estimated time remaining: %.2f seconds" % (remaining))
    else:
        pool = Pool(processes=threads)
        ndetec = np.zeros((nsep, nth, ncon))
        if projected:
            ndetec = pool.map(detec_sim_loopfit_proj, all_vars)
        elif use_cov:
            ndetec = pool.map(detec_sim_loopfit_cov, all_vars)
        else:
            ndetec = pool.map(detec_sim_loopfit, all_vars)
        ndetec = np.array(ndetec)

        # and clean up
        pool.close()

    tf = time.time()
    if tf - tic > 60:
        print("Total time elapsed:", (tf - tic) / 60.0, "mins")
    elif tf - tic <= 60:
        print("Total time elapsed:", tf - tic, "seconds")
    # nc, ns = int(ncon), int(nsep)

    # for each contrast and sep, how many detections?
    nd = ndetec
    ndetec = ndetec.sum(axis=1)
    nbtot = nsim * nth
    ndetec /= float(nbtot)
    # and as usual, ACC got the dimensions wrong (still in IDL mode)
    ndetec = np.transpose(ndetec)

    # ---------------------------------------------------------------
    #                        contour plot!
    # ---------------------------------------------------------------
    levels = [0.5, 0.9, 0.99, 0.999]
    mycols = ("k", "k", "k", "k")
    if not no_plot:
        plt.figure(0)
        plt.clf()
        contours = plt.contour(
            ndetec, levels, colors=mycols, linewidth=2, extent=[smin, smax, cmin, cmax]
        )
        plt.clabel(contours)
        plt.contourf(seps, cons, ndetec, levels, cmap=plt.cm.bone)
        plt.colorbar()
        plt.xlabel("Separation (mas)")
        plt.ylabel("Contrast Ratio")
        plt.title("Contrast Detection Limits")
        plt.draw()
        plt.show()

    data = {
        "levels": levels,
        "seps": seps,
        "angles": ths,
        "cons": cons,
        "name": cpo.name,
        "ndetec": nd,
        "nsim": nsim,
        "limits": ndetec,
    }

    if save:
        if type(save) is str:
            save_file = save
        else:
            save_file = "limit_lowc" + cpo.name + ".pick"
        print("Saving contrast limits as: " + save_file)

        with open(save_file, "w") as myf:
            pickle.dump(data, myf)

    return data


# =========================================================================
# =========================================================================


def binary_fit(cpo, p0):
    """Performs a best binary fit search for the dataset.
    -------------------------------------------------------------
    p0 is the initial guess for the parameters 3 parameter vector
    typical example would be : [100.0, 0.0, 5.0].
    returns the full solution of the least square fit:
    - soluce[0] : best-fit parameters
    - soluce[1] : covariance matrix
    -------------------------------------------------------------"""

    if np.all(cpo.t3err == 0.0):
        print("Closure phase object instance is not calibrated.\n")
        soluce = leastsq(cpo.bin_fit_residuals, p0, args=(cpo), full_output=1)
    else:

        def lmcpmodel(params, cpo=cpo):
            model = cp_model(params, cpo.u, cpo.v, cpo.wavel)
            return ((model - cpo.t3data) / cpo.t3err).ravel()

        soluce = leastsq(lmcpmodel, p0, args=cpo, full_output=1)
    cpo.covar = soluce[1]
    # to get consistent position angle measurements
    soluce[0][1] = np.mod(soluce[0][1], 360.0)
    return soluce[0:2]  # only return the best params and covariance matrix


# =========================================================================
# =========================================================================


def bin_fit_residuals(params, cpo):
    """Function for binary_fit without errorbars"""
    test = cp_model(params, cpo.u, cpo.v, cpo.wavel)
    err = cpo.t3data - test
    return err


# =========================================================================
# =========================================================================


def brute_force_chi2_grid(everything):
    """Function for multiprocessing, fills in part of the big 3d chi2 grid in a
    way that minimises repeated calculations (i.e. each model only calculated once)."""
    sim_cps = everything["sim_cps"]
    sep = everything["sep"]
    nth = everything["nth"]
    ncon = everything["ncon"]
    nsim = everything["nsim"]
    chi2 = np.zeros((nth, ncon, nsim))
    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):
            mod_cps = cp_model(
                [sep, th, con], everything["u"], everything["v"], everything["wavel"]
            )
            for i in range(nsim):
                chi2[j, k, i] = np.sum(
                    ((sim_cps[:, :, i] - mod_cps) / everything["error"]) ** 2
                )
    return chi2


# =========================================================================
# =========================================================================


# def brute_force_chi2_grid_proj(everything):
#     '''
#     Function for multiprocessing, fills in part of the big 3d chi2 grid in a
#     way that minimises repeated calculations (i.e. each model only calculated once).
#     This version handles projected closure phases
#     '''
#     sim_cps = everything['sim_cps']
#     sep = everything['sep']
#     nth = everything['nth']
#     ncon = everything['ncon']
#     nsim = everything['nsim']
#     n_runs = everything['n_runs']
#     n_clps = everything['n_clps']
#     chi2 = np.zeros((nth, ncon, nsim))
#     for j, th in enumerate(everything['ths']):
#         for k, con in enumerate(everything['cons']):
#             mod_cps = cp_model([sep, th, con], everything['u'],
#                                everything['v'], everything['wavel'])
#             # project and take independent clps
#             # THIS NEEDS TO BE FIXED
#             proj_bin_cps = project_cps(mod_cps, proj)
#             mod_cps = np.reshape(
#                 mod_cps, (len(mod_cps)/n_runs, n_runs), order='F')
#             proj_bin_cp = np.transpose(everything['proj']).dot(mod_cps)
#             proj_bin_cp = proj_bin_cp[0:n_clps, :]
#             proj_bin_cp = proj_bin_cp.ravel()
#             for i in range(nsim):
#                 chi2[j, k, i] = np.sum(
#                     ((sim_cps[:, :, i]-proj_bin_cp) / everything['error'])**2)
#     return chi2

# =========================================================================
# =========================================================================


def lmfit(everything):
    """Unpacks a cpo and calls the proper binary L-M fitting function"""
    this_cpo = everything["sim_cpo"]
    best_chi2_params = everything["best_chi2_params"]
    [new_params, cov] = binary_fit(this_cpo, best_chi2_params)
    return new_params


# =========================================================================
# =========================================================================


# def brute_force_detec_limits(cpo, nsim=100, nsep=32, nth=20, ncon=32, smin='Default', smax='Default',
#                              cmin=10., cmax=500., extra_error=0, threads=0, save=False, include_cov=False, icpo=False, projected=False,
#                              err_scale=1., lm_fit=True):
#     '''uses a Monte Carlo simulation to establish contrast-separation
#     detection limits given an array of standard deviations per closure phase.

#     Because different separation-contrast grid points are entirely
#     separate, this task is embarrassingly parallel. If you want to
#     speed up the calculation, use multiprocessing with a threads
#     argument equal to the number of available cores.

#     Make nseps a multiple of threads! This uses the cores most efficiently.

#     Hyperthreading (2x processes per core) in my experience gets a ~20%
#     improvement in speed.

#     Written by F. Martinache and B. Pope.

#     This version was modified by ACC to use a brute force
#     chi2 grid that emulates the idl program binary_grid.pro.

#     lm_fit option allows an extra step involving a levenberg-marquadt fit at each point,
#     to ensure that each set of best parameters is the actual minimum and reduce the effect of
#     the grid size.
#     '''

#     # ------------------------
#     # first, load your data!
#     # ------------------------

#     if projected is True:
#         proj = cpo.proj
#         cpo.t3err = np.sqrt(cpo.t3err**2+extra_error**2)
#         error = cpo.t3err.ravel()*err_scale
#         n_clps = cpo.n_good
#         n_runs = cpo.n_runs
#     elif icpo is True:
#         proj = []
#         cpo.t3err = np.sqrt(cpo.t3err**2+extra_error**2)
#         error = cpo.t3err.ravel()*err_scale
#         n_clps = cpo.n_clps
#         n_runs = cpo.n_runs
#     else:
#         proj = []
#         cpo.t3err = np.sqrt(cpo.t3err**2+extra_error**2)
#         error = cpo.t3err.ravel()*err_scale
#         n_clps = cpo.ndata
#         n_runs = 1

#     nwav = cpo.wavel.size
#     ndata = cpo.u.shape[0]
#     u = np.repeat(np.resize(cpo.u, [ndata, 3, 1]), nwav, 2).ravel()
#     v = np.repeat(np.resize(cpo.v, [ndata, 3, 1]), nwav, 2).ravel()
#     wavel = np.repeat(
#         np.repeat(np.resize(cpo.wavel, [1, 1, nwav]), ndata, 0), 3, 1).ravel()

#     ndata = n_clps*n_runs

#     w = np.array(np.sqrt(u**2 + v**2))/wavel

#     if smin == 'Default':
#         smin = rad2mas(1./4/np.max(w))

#     if smax == 'Default':
#         smax = rad2mas(1./np.min(w))

#     # ------------------------
#     # initialise Monte Carlo
#     # ------------------------

#     seps = smin + (smax-smin) * np.linspace(0, 1, nsep)
#     ths = 0.0 + 360.0 * np.linspace(0, 1, nth)
#     cons = cmin + (cmax-cmin) * np.linspace(0, 1, ncon)

#     if include_cov is True:
#         # use full multivariate gaussian with sample covariance to get rands
#         rands = np.zeros((n_clps*n_runs, nwav, nsim))
#         for run in range(n_runs):
#             rands[n_clps*(run):n_clps*(run+1), 0, :] = np.transpose(
#                 np.random.multivariate_normal(np.zeros(n_clps), cpo.sample_cov, (nsim)))
#     else:
#         rands = np.random.randn(ndata, nwav, nsim)
#     error = np.reshape(error, [ndata, nwav])
#     sim_cps = np.reshape(rands*error[:, :, np.newaxis], (ndata, nwav, nsim))

#     sim_cpos = []
#     for ix in range(nsim):
#         this_cpo = copy.deepcopy(cpo)
#         this_cpo.t3data = sim_cps[:, :, ix]
#         sim_cpos.append(this_cpo)

#     # ------------------------
#     # Run Monte Carlo
#     # ------------------------
#     tic = time.time()  # start the clock
#     best_params = []
#     if threads == 0:
#         toc = time.time()
#         chi2 = np.zeros((nsep, nth, ncon, nsim))
#         print('Starting big loop over separations')
#         for ix in range(nsep):
#             everything = {'sep': seps[ix], 'cons': cons, 'ths': ths, 'nsim': nsim,
#                           'nsep': nsep, 'ncon': ncon, 'nth': nth, 'u': cpo.u, 'v': cpo.v,
#                           'sim_cps': sim_cps, 'error': error, 'wavel': cpo.wavel, 'n_runs': n_runs,
#                           'n_clps': n_clps, 'proj': proj}
#             if projected:
#                 chi2[ix, :, :, :] = brute_force_chi2_grid_proj(everything)
#             else:
#                 chi2[ix, :, :, :] = brute_force_chi2_grid(everything)
#             if (ix % 5) == 0:
#                 tc = time.time()
#                 print('Done', ix, '. Time taken:', (tc-toc), 'seconds')
#                 toc = tc
#     else:
#         all_vars = []
#         print('Making big array')
#         for ix in range(nsep):
#             everything = {'sep': seps[ix], 'cons': cons, 'ths': ths, 'nsim': nsim,
#                           'nsep': nsep, 'ncon': ncon, 'nth': nth, 'u': cpo.u, 'v': cpo.v,
#                           'sim_cps': sim_cps, 'error': error, 'wavel': cpo.wavel, 'n_runs': n_runs,
#                           'n_clps': n_clps, 'proj': proj}
#             all_vars.append(everything)
#         print('Made big array for loop')
#         pool = Pool(processes=threads)
#         if projected:
#             chi2 = pool.map(brute_force_chi2_grid_proj, all_vars)
#         else:
#             chi2 = pool.map(brute_force_chi2_grid, all_vars)
#         pool.close()
#         chi2 = np.array(chi2)
#     tf = time.time()
#     if tf-tic > 60:
#         print('Total time elapsed:', (tf-tic)/60., 'mins')
#     elif tf-tic <= 60:
#         print('Total time elapsed:', tf-tic, 'seconds')

#     # ------------------------
#     # Do L-M fit to get actual best fit params
#     # ------------------------
#     print('Finding best parameters')
#     best_params = np.zeros((nsim, 3))
#     best_chi2_params = np.zeros((nsim, 3))
#     for ix in range(nsim):
#         b_params_ix = np.where(chi2[:, :, :, ix] == np.amin(chi2[:, :, :, ix]))
#         b_params = [seps[b_params_ix[0][0]],
#                     ths[b_params_ix[1][0]], cons[b_params_ix[2][0]]]
#         best_chi2_params[ix, :] = np.array(b_params)
#     print('Found best parameters')

#     if lm_fit:
#         print('Starting L-M fits.')
#         tic = time.time()  # start the clock
#         if threads == 0:
#             for ix in range(nsim):
#                 this_cpo = sim_cpos[ix]
#                 [new_params, cov] = binary_fit(
#                     this_cpo, best_chi2_params[ix, :])
#                 if ix % 50 == 0:
#                     print('Done', ix)
#                 best_params[ix, :] = new_params
#         else:
#             all_vars = []
#             for ix in range(nsim):
#                 # print('Making big array:',ix,'of',nsep)
#                 everything = {
#                     'sim_cpo': sim_cpos[ix], 'best_chi2_params': best_chi2_params[ix, :]}
#                 all_vars.append(everything)
#             print('Using', threads, 'cores')
#             pool = Pool(processes=threads)
#             best_params = pool.map(lmfit, all_vars)
#             pool.close()

#         tf = time.time()
#         if tf-tic > 60:
#             print('Total time elapsed:', (tf-tic)/60., 'mins')
#         elif tf-tic <= 60:
#             print('Total time elapsed:', tf-tic, 'seconds')
#     else:
#         best_params = best_chi2_params

#     # ------------------------
#     # Collect and organise
#     # ------------------------

#     ndetec = np.zeros((ncon, nsep))
#     # nc, ns = int(ncon), int(nsep)

#     # collect them
#     for ix in range(nsim):
#         sep = best_params[ix][0]
#         con = best_params[ix][2]
#         sep_ix = np.where(abs(seps-sep) == np.amin(abs(seps-sep)))
#         con_ix = np.where(abs(cons-con) == np.amin(abs(cons-con)))
#         ndetec[sep_ix, con_ix] += 1
#     # Take the cumulative sum over contrast ratio at each sep
#     cumsum_detec = ndetec.cumsum(axis=1)
#     # turn into %
#     maxdetec = np.amax(cumsum_detec, axis=1)
#     ndetec = 0*cumsum_detec
#     for ix in range(nsep):
#         if maxdetec[ix] == 0:
#             print('No sims for sep '+str(seps[ix])+'mas')
#         else:
#             print(str(maxdetec[ix])+" in "+str(seps[ix])+" bin.")
#             ndetec[ix, :] = cumsum_detec[ix, :]/maxdetec[ix]

#     ndetec = 1-ndetec
#     # Axes were wrong way around (I blame IDL)
#     ndetec = np.transpose(ndetec)

#     # ---------------------------------------------------------------
#     #                        contour plot!
#     # ---------------------------------------------------------------
#     levels = [0., 0.9, 0.99, 0.999]
#     mycols = ('k', 'k', 'k', 'k')

#     plt.figure(0)
#     contours = plt.contour(ndetec, levels, colors=mycols, linewidth=2,
#                            extent=[smin, smax, cmin, cmax])
#     plt.clabel(contours)
#     plt.contourf(seps, cons, ndetec, levels, cmap=plt.cm.bone)
#     plt.colorbar()
#     plt.xlabel('Separation (mas)')
#     plt.ylabel('Contrast Ratio')
#     plt.title('Contrast Detection Limits')
#     plt.draw()
#     plt.show()

#     data = {'levels': levels,
#             'seps': seps,
#             'angles': ths,
#             'cons': cons,
#             'name': cpo.name,
#             'cumsum': cumsum_detec,
#             'best_params': best_params,
#             'limits': ndetec}

#     if save is True:
#         file = 'limit_lowc'+cpo.name+'.pick'
#         print(file)

#         myf = open(file, 'w')
#         pickle.dump(data, myf)
#         myf.close()

#     return data


# =========================================================================
# =========================================================================


def chi2_grid(everything):
    """Function for multiprocessing, does 2d chi2 grid for coarse_grid"""
    cpo = everything["cpo"]
    data_cp = cpo.t3data
    chi2 = np.zeros((len(everything["ths"]), len(everything["cons"])))
    sep = everything["sep"]
    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):
            mod_cps = cp_model([sep, th, con], cpo.u, cpo.v, cpo.wavel)
            chi2[j, k] = np.sum(
                ((data_cp.ravel() - mod_cps.ravel()) / cpo.t3err.ravel()) ** 2
            )
    return chi2


# =========================================================================
# =========================================================================


def chi2_grid_cov(everything):
    """Function for multiprocessing, does 2d chi2 grid for coarse_grid, using
    the covariance matrix instead of uncertainties"""
    cpo = everything["cpo"]
    data_cp = cpo.t3data
    chi2 = np.zeros((len(everything["ths"]), len(everything["cons"])))
    sep = everything["sep"]
    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):
            mod_cps = cp_model([sep, th, con], cpo.u, cpo.v, cpo.wavel)
            resids = data_cp.ravel() - mod_cps.ravel()

            chi2[j, k] = resids.transpose().dot(cpo.cov_inv).dot(resids)
    return chi2


# =========================================================================
# =========================================================================


def chi2_grid_proj(everything):
    """Function for multiprocessing, does 2d chi2 grid for coarse_grid, with projected data"""
    cpo = everything["cpo"]
    chi2 = np.zeros((len(everything["ths"]), len(everything["cons"])))
    sep = everything["sep"]
    for j, th in enumerate(everything["ths"]):
        for k, con in enumerate(everything["cons"]):
            mod_cps = cp_model([sep, th, con], cpo.u, cpo.v, cpo.wavel)
            proj_mod_cps = project_cps(mod_cps, cpo.proj)
            chi2[j, k] = np.sum(
                ((cpo.proj_t3data - proj_mod_cps) / cpo.proj_t3err) ** 2
            )
    return chi2


def coarse_grid(
    cpo,
    nsep=32,
    nth=20,
    ncon=32,
    smin="Default",
    smax="Default",
    cmin=10.0,
    cmax=500.0,
    threads=0,
    projected=False,
    err_scale=1.0,
    extra_error=0.0,
    thmin=0.0,
    thmax=360.0,
    use_cov=False,
    plot=True,
    verbose=True,
):
    """Does a coarse grid search for the best fit parameters. This is
    helpful for finding a good initial point for hammer or multinest.

    Because different separation-contrast grid points are entirely
    separate, this task is embarrassingly parallel. If you want to
    speed up the calculation, use multiprocessing with a threads
    argument equal to the number of available cores.

    Make nseps a multiple of threads! This uses the cores most efficiently.

    Hyperthreading (2x processes per core) in my experience gets a ~20%
    improvement in speed.

    Written by A Cheetham, with some parts stolen from other pysco/pymask routines.

    use_cov : use inverse covariance matrix stored in cpo.cov_inv instead of uncertainties

    """

    # ------------------------
    # first, load your data!
    # ------------------------
    try:
        if projected is True:
            ndata = cpo.n_runs * cpo.n_good
        else:
            ndata = cpo.ndata
    except AttributeError:
        return None

    u, v = cpo.u, cpo.v

    cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
    cpo.t3err *= err_scale

    wavel = cpo.wavel

    w = np.array(np.sqrt(u ** 2 + v ** 2))

    if smin == "Default":
        smin = rad2mas(1.0 / 4 / np.max(w / np.min(wavel)))

    if smax == "Default":
        smax = rad2mas(1.0 / np.min(w / np.max(wavel)))

    # ------------------------
    # initialise grid params
    # ------------------------

    seps = smin + (smax - smin) * np.linspace(0, 1, nsep)
    ths = np.linspace(thmin, thmax, num=nth)
    cons = cmin + (cmax - cmin) * np.linspace(0, 1, ncon)

    # ------------------------
    # Calculate chi2 at each point
    # ------------------------
    tic = time.time()  # start the clock
    chi2 = np.zeros((nsep, nth, ncon))

    if threads == 0:
        toc = time.time()
        for ix, sep in enumerate(seps):
            everything = {
                "sep": seps[ix],
                "cons": cons,
                "ths": ths,
                "cpo": cpo,
                "ix": ix,
            }
            if projected:
                chi2[ix, :, :] = chi2_grid_proj(everything)
            elif use_cov:
                chi2[ix, :, :] = chi2_grid_cov(everything)
            else:
                chi2[ix, :, :] = chi2_grid(everything)
            if (ix % 50) == 0:
                tc = time.time()
                if verbose:
                    print("Done", ix, ". Time taken:", (tc - toc), "seconds")
                toc = tc
    else:
        all_vars = []
        for ix in range(nsep):
            everything = {
                "sep": seps[ix],
                "cons": cons,
                "ths": ths,
                "cpo": cpo,
                "ix": ix,
            }
            all_vars.append(everything)
        pool = Pool(processes=threads)
        if projected:
            chi2 = pool.map(chi2_grid_proj, all_vars)
        elif use_cov:
            chi2 = pool.map(chi2_grid_cov, all_vars)
        else:
            chi2 = pool.map(chi2_grid, all_vars)
    tf = time.time()
    if tf - tic > 60:
        if verbose:
            print("Total time elapsed:", (tf - tic) / 60.0, "mins")
    elif tf - tic <= 60:
        if verbose:
            print("Total time elapsed:", tf - tic, "seconds")
    chi2 = np.array(chi2)
    best_ix = np.where(chi2 == np.amin(chi2))
    # If the best chi2 is at more than one location, take the first.
    best_params = [seps[best_ix[0][0]], ths[best_ix[1][0]], cons[best_ix[2][0]]]
    best_params = np.array(np.array(best_params).ravel())
    print("\nMaximum likelihood estimation (χ2=%2.1f):" % (np.amin(chi2) / (ndata)))
    print("-------------------------------")
    print("Separation = %2.2f mas" % best_params[0])
    print("PA = %2.2f deg" % best_params[1])
    print(
        "Contrast Ratio = %2.1f (%2.1f mag)\n"
        % (best_params[2], 2.5 * np.log10(best_params[2]))
    )
    # ---------------------------------------------------------------
    #   sum over each variable so we can visualise it all
    # ---------------------------------------------------------------
    temp_chi2 = ndata * chi2 / np.amin(chi2)
    like = np.exp(-(temp_chi2 - ndata) / 2)
    sep_pa = np.sum(like, axis=2)
    sep_crat = np.sum(like, axis=1)
    pa_crat = np.sum(like, axis=0)

    # ---------------------------------------------------------------
    #                        contour plot!
    # ---------------------------------------------------------------
    if plot:
        vmax = np.max([sep_pa.max(), sep_crat.max(), pa_crat.max()])
        plt.figure(figsize=(6, 5))
        plt.subplot(2, 2, 1)
        plt.imshow(
            sep_pa,
            extent=[np.amin(ths), np.amax(ths), np.amin(seps), np.amax(seps)],
            cmap="YlGnBu_r",
            vmax=vmax,
            aspect="auto",
        )
        plt.colorbar()
        plt.ylabel("Separation [mas]")
        plt.xlabel("PA [deg]")
        plt.subplot(2, 2, 2)
        plt.imshow(
            sep_crat,
            extent=[np.amin(cons), np.amax(cons), np.amin(seps), np.amax(seps)],
            cmap="YlGnBu_r",
            vmax=vmax,
            aspect="auto",
        )
        plt.colorbar()
        plt.gca().tick_params(axis="x", label1On=False)
        plt.gca().tick_params(axis="y", label1On=False)
        plt.subplot(2, 2, 4)
        plt.imshow(
            pa_crat,
            extent=[np.amin(cons), np.amax(cons), np.amin(ths), np.amax(ths)],
            cmap="YlGnBu_r",
            vmax=vmax,
            aspect="auto",
        )
        plt.colorbar()
        plt.ylabel("PA [deg]")
        plt.xlabel("Contrast Ratio")
        plt.subplots_adjust(
            top=0.974, bottom=0.092, left=0.097, right=0.99, hspace=0.1, wspace=0.05
        )
        plt.show(block=False)

    data = {
        "chi2": chi2,
        "like": like,
        "seps": seps,
        "ths": ths,
        "cons": cons,
        "best": best_params,
    }
    return data


def test_significance(cpo, params, projected=False):
    """Tests whether a certain set of binary parameters is a better fit than a
    single star model"""
    # first, calculate the null hypothesis chi squared
    if projected:
        t3data = cpo.proj_t3data
        t3err = cpo.proj_t3err
    else:
        t3data = cpo.t3data
        t3err = cpo.t3err
        u = cpo.u
        v = cpo.v
        wavel = cpo.wavel

    null_chi2 = np.sum((t3data.ravel() / t3err.ravel()) ** 2)
    mod_cps = cp_model(params, u, v, wavel)
    if projected:
        mod_cps = np.reshape(mod_cps, (cpo.n_clps, cpo.n_runs), order="F")
        mod_cps = (np.transpose(cpo.proj).dot(mod_cps))[: len(t3data), :]
    bin_chi2 = np.sum(((t3data.ravel() - mod_cps.ravel()) / t3err.ravel()) ** 2)
    print("Chi2_null:", null_chi2)
    print("Chi2_bin:", bin_chi2)
    print("Significance:", np.sqrt(null_chi2 - bin_chi2), "=sqrt(chi2_null-chi2_bin)")
    return [null_chi2, bin_chi2]


def multiple_companions_hammer(
    cpo,
    ivar=[[50.0, 0.0, 2.0], [50.0, 90.0, 2.0]],
    ndim="Default",
    nwalcps=50,
    plot=False,
    projected=False,
    niters=1000,
    threads=1,
    model="constant",
    sep_prior=None,
    pa_prior=None,
    crat_prior=None,
    err_scale=1.0,
    extra_error=0.0,
):
    """Implementation of emcee, the MCMC Hammer, for closure phase
    fitting to multiple companions. Requires a closure phase object cpo, and is best called with
    ivar chosen to be near the peak - it can fail to converge otherwise.
    See cp_model for details!
    Prior ranges introduce a flat (tophat) prior between the two values specified
    NOTE: This doesn't work with the wavelength model yet!"""

    import emcee

    ivar = np.array(ivar)  # initial parameters for model-fit
    if ndim == "Default":
        ndim = len(ivar)

    if ivar.ndim != 1:
        ncomp = ivar.shape[0]
        ndim = ivar.size
        nparam = ndim / ncomp
    else:
        ncomp = 1
        nparam = ndim

    cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
    cpo.t3err *= err_scale

    print("Fitting for:", ncomp, "companions")

    p0 = [
        ivar.ravel() + 0.1 * ivar.ravel() * np.random.rand(ndim) for i in range(nwalcps)
    ]  # initialise walcps in a ball
    print("Running emcee now!")

    t0 = time.time()
    if projected is False:
        sampler = emcee.EnsembleSampler(
            nwalcps,
            ndim,
            cp_loglikelihood_multiple,
            args=[cpo.u, cpo.v, cpo.wavel, cpo.t3data, cpo.t3err, model, ncomp],
            threads=threads,
        )
    else:
        print("I havent coded this yet...")
        raise NameError("cp_loglikelihood_multiple_proj doesnt exist")

    sampler.run_mcmc(p0, niters)
    tf = time.time()

    print("Time elapsed =", tf - t0, "s")

    chain = sampler.flatchain

    seps = np.zeros((chain.shape[0], ncomp))
    ths = np.zeros((chain.shape[0], ncomp))
    cs = np.zeros((chain.shape[0], ncomp, nparam - 2))
    for ix in range(ncomp):
        seps[:, ix] = chain[:, ix * nparam]
        ths[:, ix] = chain[:, ix * nparam + 1]
        cs[:, ix, :] = chain[:, ix * nparam + 2 : (ix + 1) * nparam]

    # Now introduce the prior, by ignoring values outside of the range
    if sep_prior is not None:
        for c_ix in range(ncomp):
            wh = (seps[:, c_ix] >= sep_prior[0]) & (seps[:, c_ix] <= sep_prior[1])
            seps = seps[wh, :]
            ths = ths[wh, :]
            cs = cs[wh, :]
    if pa_prior is not None:
        for c_ix in range(ncomp):
            wh = (ths[:, c_ix] >= pa_prior[0]) & (ths[:, c_ix] <= pa_prior[1])
            seps = seps[wh, :]
            ths = ths[wh, :]
            cs = cs[wh, :, :]
    if crat_prior is not None:
        for c_ix in range(ncomp):
            for ix in range(nparam - 2):
                wh = (cs[:, c_ix, ix] >= crat_prior[0]) & (
                    cs[:, c_ix, ix] <= crat_prior[1]
                )
                seps = seps[wh]
                ths = ths[wh]
                cs = cs[wh, :, :]

    # Now put the "good" values back into chain for later
    chain = np.zeros((seps.shape[0], nparam * ncomp))
    for ix in range(ncomp):
        chain[:, ix * nparam] = seps[:, ix]
        chain[:, ix * nparam + 1] = ths[:, ix]
        chain[:, ix * nparam + 2 : (ix + 1) * nparam] = cs[:, ix, :]

    if plot is True:
        plt.clf()

    # Loop over number of companions to calculate best params and do plots
    meanseps = []
    dseps = []
    meanths = []
    dths = []
    meancs = []
    dcs = []
    for c_ix in range(ncomp):
        meansep = np.mean(seps[:, c_ix])
        dsep = np.std(seps[:, c_ix])

        meanth = np.mean(ths[:, c_ix])
        dth = np.std(ths[:, c_ix])

        meanc = np.mean(cs[:, c_ix, :], axis=0)
        dc = np.std(cs[:, c_ix], axis=0)

        print("Companion #" + str(c_ix))
        print("    Separation", meansep, "pm", dsep, "mas")
        print("    Position angle", meanth, "pm", dth, "deg")
        print("    Contrast", meanc[0], "pm", dc[0])

        meanseps.append(meansep)
        dseps.append(dsep)
        meanths.append(meanth)
        dths.append(dth)
        meancs.append(meanc)
        dcs.append(dc)

        if model == "linear":
            print("    Contrast2", meanc[1], "pm", dc[1])
            extra_pars = ["Contrast "]
            extra_dims = ["Ratio"]
        elif model == "free":
            extra_pars = np.repeat("Contrast", ndim - 2)
            extra_dims = np.repeat("Ratio", ndim - 2)
        else:
            extra_pars = "None"
            extra_dims = "None"

        if plot is True:
            paramnames = ["Separation", "Position Angle", "Contrast"]
            paramdims = ["(mas)", "(deg)", "Ratio"]
            for ix, par in enumerate(extra_pars):
                paramnames.append(par)
                paramdims.append(extra_dims[ix])

            for i in range(nparam):
                if i < 3:
                    plt.figure(i)
                    if c_ix == 0:
                        plt.clf()
                plt.hist(chain[:, c_ix * nparam + i], niters / 5, histtype="step")
                plt.title(paramnames[i])
                plt.ylabel("Counts")
                plt.xlabel(paramnames[i] + " " + paramdims[i])

            plt.show()

    data = {
        "sep": meanseps,
        "delsep": dseps,
        "pa": meanths,
        "delpa": dths,
        "con": meancs,
        "delcon": dcs,
        "chain": sampler.chain,
        "seps": seps,
    }
    # and clean up
    sampler.pool.terminate()
    return data


# =========================================================================
# =========================================================================


def hammer_spectrum(
    cpo,
    params,
    ivar=[1.53],
    nwalcps=50,
    plot=False,
    model="free",
    niters=1000,
    threads=1,
    crat_prior=None,
    err_scale=1.0,
    extra_error=0.0,
):
    """ACC's modified version of hammer that allows fitting to a spectrum only, by
    fixing the position angle and separation of the companion. Made for HD142527
    Requires a closure phase object cpo, and is best called with
    ivar chosen to be near the peak - it can fail to converge otherwise.
    Also allows fitting of a contrast vs wavlength model. See cp_model for details!
    Prior ranges introduce a flat (tophat) prior between the two values specified.

    Ndim is the number of d.o.f in the spectral channels you want to fit to.
    It only works if ndim=nwav at the moment. Whoops. Otherwise need a polynomial model for cp_model"""

    import emcee

    [sep, pa] = params

    ivar = np.array(ivar)  # initial parameters for spectrum fit
    ndim = len(ivar)
    if ndim != 37:
        print(
            "This is still a work in progress, and only works for 37 channel data (GPI default)"
        )

    # initialise walcps in a ball
    p0 = [ivar + 0.1 * ivar * np.random.rand(ndim) for i in range(nwalcps)]

    p0 = np.array(p0)
    params = np.array(params)
    cpo.t3err = np.sqrt(cpo.t3err ** 2 + extra_error ** 2)
    cpo.t3err *= err_scale

    print("Running emcee now!")

    t0 = time.time()
    sampler = emcee.EnsembleSampler(
        nwalcps,
        ndim,
        cp_loglikelihood_spectrum,
        args=[params, cpo.u, cpo.v, cpo.wavel, cpo.t3data, cpo.t3err, model],
        threads=threads,
    )
    sampler.run_mcmc(p0, niters)
    # x=cp_loglikelihood_spectrum(ivar,params,cpo.u,cpo.v,cpo.wavel,cpo.t3data,cpo.t3err,model)
    tf = time.time()

    print("Time elapsed =", tf - t0, "s")

    chain = sampler.flatchain
    cs = chain

    # Now introduce the prior, by ignoring values outside of the range
    if crat_prior is not None:
        for ix in range(ndim - 2):
            c = cs[:, ix]
            wh = (c >= crat_prior[0]) & (c <= crat_prior[1])
            cs = cs[wh, :]

    # check that there are still some points left!
    if cs.size > 0:
        chain = np.zeros((len(cs), ndim))
        chain = cs
    else:
        print("WARNING: Your priors eliminated all points!")

    meanc = np.mean(cs, axis=0)
    dc = np.std(cs, axis=0)
    print("Mean contrast", np.mean(meanc), "pm", np.mean(dc))
    for ix in range(ndim):
        print("Channel", ix, ":", meanc[ix], "pm", dc[ix])

    if plot is True:

        plt.clf()

        for i in range(ndim):
            plt.hist(chain[:, i], niters / 10, histtype="step")
            plt.title("Contrast")
            plt.ylabel("Counts")
            plt.xlabel("Contrast Ratio")

        plt.show()

    data = {
        "con": meanc,
        "delcon": dc,
        "chain": sampler.chain,
        "posterior": chain,
        "wavelength": cpo.wavel,
    }
    # and clean up
    sampler.pool.terminate()
    return data


# =========================================================================
# =========================================================================


def cp_loglikelihood_spectrum(
    spec_params, bin_params, u, v, wavel, t3data, t3err, model="free"
):
    """Calculate loglikelihood for closure phase data, with fixed sep,
    pa and con so you can measure a spectrum. Used in hammer_spectrum."""
    params = np.concatenate((bin_params, spec_params))
    cps = cp_model(params, u, v, wavel, model=model)
    chi2 = np.sum(((t3data.ravel() - cps.ravel()) / t3err.ravel()) ** 2)
    loglike = -chi2 / 2
    return loglike


# =========================================================================
# =========================================================================


def find_extra_error(
    params, cpo, err_scale=1.0, dof="Default", model="constant", projected=False
):
    """Finds the extra error needed for reduced chi squared to be 1. This is
    done by trying for many values of extra error and then interpolating.
    If dof is not set, then the degrees of freedom are taken as cpo.t3data.size."""

    model_cps = cp_model(params, cpo.u, cpo.v, cpo.wavel, model=model)
    if projected:
        model_cps = project_cps(model_cps, cpo.proj)
        data_cps = cpo.proj_t3data
        cp_err = cpo.proj_t3err
    else:
        data_cps = cpo.t3data
        cp_err = cpo.t3err

    cp_resids = data_cps - model_cps

    print("Max resid:" + str(np.max(np.abs(cp_resids))))

    n = 1000  # number of extra errors to try.
    chis_null = np.zeros(n)
    chis_bin = np.zeros(n)

    extra_errors = np.logspace(-2, 2, num=n)

    for ix, err in enumerate(extra_errors):

        # Calculate the closure phase uncertainties for this amount of extra_error
        t3err = np.sqrt(cp_err ** 2 + err ** 2)
        t3err *= err_scale

        chis_bin[ix] = np.sum(cp_resids ** 2 / t3err ** 2)
        chis_null[ix] = np.sum(data_cps ** 2 / t3err ** 2)

    # scale to red.chi2
    if dof == "Default":
        dof = data_cps.size

    chis_bin /= dof - 3.0  # -3 due to the fit
    chis_null /= dof

    print("Binary Chi2 with no extra error: " + str(np.max(chis_bin)))

    # now find where it goes to 1. with interpolation
    func_bin = interp.interp1d(chis_bin, extra_errors)
    func_null = interp.interp1d(chis_null, extra_errors)
    try:
        err_bin = func_bin(1.0)
    except Exception:
        if np.min(chis_bin) < 1.0:
            print("Binary Reduced Chi2 less than 1")
        else:
            print("Problem trying to calculate extra error.")
        err_bin = 0.0

    try:
        err_null = func_null(1.0)
    except Exception:
        if np.min(chis_null) < 1.0:
            print("Null Reduced Chi2 less than 1")
        else:
            print("Problem trying to calculate extra error.")
        err_null = 0.0

    print("Extra error needed:")
    print("Binary:", err_bin)
    print("Null  :", err_null)

    # Diagnostic plots
    cp_err_plot = np.sqrt(cp_err ** 2 + err_bin ** 2)

    plt.clf()
    plt.subplot(211)
    plt.errorbar(model_cps.ravel(), data_cps.ravel(), cp_err_plot.ravel(), fmt="o")
    plt.xlabel("Model CLP (deg)")
    plt.ylabel("Measured CLP (deg)")
    plt.subplot(212)
    plt.errorbar(model_cps.ravel(), cp_resids.ravel(), cp_err_plot.ravel(), fmt="o")
    plt.xlabel("Model CLP (deg)")
    plt.ylabel("Residual CLP (deg)")
    plt.tight_layout()

    return chis_bin, extra_errors


# =========================================================================
# =========================================================================


# def multiple_companions_nest(cpo, paramlimits, n_comp=2., resume=False, eff=0.3, multi=True,
#                              err_scale=1., extra_error=0., plot=True, npoints=1000, logcrat_prior=True):
#     '''Default implementation of a MultiNest fitting routine for closure
#     phase data. Requires a closure phase cpo object, parameter limits and
#     sensible keyword arguments for the multinest parameters.

#     This function does very naughty things creating functions inside this
#     function because PyMultiNest is very picky about how you pass it
#     data.

#     Optional parameter eff tunes sampling efficiency, and multi toggles multimodal
#     nested sampling on and off. Turning off multimodal sampling results in a speed
#     boost of ~ 20-30%.

#     logcrat_prior allows the prior for the contrast ratios to be flat in logarithmic space

#     paramlimits is needed here, and must be a list with length = ncomp,
#     with paramlimits[companion_number]=[minsep,maxsep,minpa,maxpa,mincrat,maxcrat]
#     '''
#     import pymultinest

#     if logcrat_prior:
#         def myprior(cube, ndim, n_params, paramlimits=paramlimits, kpo=0, n_comp=n_comp):
#             for comp_ix in range(0, n_comp):
#                 cube[3*comp_ix+0] = cube[3*comp_ix+0] * \
#                     (paramlimits[comp_ix][1] - paramlimits[comp_ix]
#                      [0])+paramlimits[comp_ix][0]
#                 cube[3*comp_ix+1] = cube[3*comp_ix+1] * \
#                     (paramlimits[comp_ix][3] - paramlimits[comp_ix]
#                      [2])+paramlimits[comp_ix][2]
#                 cube[3*comp_ix+2] = 10**(cube[3*comp_ix+2]*(np.log10(paramlimits[comp_ix][5]) -
#                                                             np.log10(paramlimits[comp_ix][4]))+np.log10(paramlimits[comp_ix][4]))

#     else:
#         def myprior(cube, ndim, n_params, paramlimits=paramlimits, kpo=0, n_comp=n_comp):
#             n_comp = ndim/3
#             for comp_ix in range(0, n_comp):
#                 cube[3*comp_ix+0] = cube[3*comp_ix+0] * \
#                     (paramlimits[comp_ix][1] - paramlimits[comp_ix]
#                      [0])+paramlimits[comp_ix][0]
#                 cube[3*comp_ix+1] = cube[3*comp_ix+1] * \
#                     (paramlimits[comp_ix][3] - paramlimits[comp_ix]
#                      [2])+paramlimits[comp_ix][2]
#                 cube[3*comp_ix+2] = cube[3*comp_ix+2] * \
#                     (paramlimits[comp_ix][5] - paramlimits[comp_ix]
#                      [4])+paramlimits[comp_ix][4]

#     def myloglike(cube, ndim, n_params):
#         loglike = cp_loglikelihood_multiple(
#             cube, cpo.u, cpo.v, cpo.wavel, cpo.t3data, cpo.t3err, ncomp=n_comp)
#         return loglike

#     # Add in the extra error and scale it
#     cpo.t3err = np.sqrt(cpo.t3err**2+extra_error**2)
#     cpo.t3err *= err_scale

#     parameters = n_comp*['Separation', 'Position Angle', 'Contrast']
#     n_params = 3*n_comp

#     # Check that the "chains" directory exists (which multinest needs)
#     if os.path.exists(os.getcwd()+'/chains/') is False:
#         os.mkdir(os.getcwd()+'/chains/')

#     tic = time.time()  # start timing

#     # ---------------------------------
#     # now run MultiNest!
#     # ---------------------------------

#     pymultinest.run(myloglike, myprior, n_comp*3, wrapped_params=[1], resume=resume, verbose=True,
#                     sampling_efficiency=eff, multimodal=multi, n_iter_before_update=1000,
#                     n_live_points=npoints)
#     # let's analyse the results
#     a = pymultinest.Analyzer(n_params=n_comp*3)
#     s = a.get_stats()

#     toc = time.time()

#     if toc-tic < 60.:
#         print('Time elapsed =', toc-tic, 's')
#     else:
#         print('Time elapsed =', (toc-tic)/60., 'mins')

#     # json.dump(s, file('%s.json' % a.outputfiles_basename, 'w'), indent=2)
#     print()
#     print("-" * 30, 'ANALYSIS', "-" * 30)
#     print("Global Evidence:\n\t%.15e +- %.15e" %
#           (s['global evidence'], s['global evidence error']))
#     params = s['marginals']

#     for comp_ix in range(n_comp):
#         bestsep = params[comp_ix*3+0]['median']
#         seperr = params[comp_ix*3+0]['sigma']

#         bestth = params[comp_ix*3+1]['median']
#         therr = params[comp_ix*3+1]['sigma']

#         bestcon = params[comp_ix*3+2]['median']
#         conerr = params[comp_ix*3+2]['sigma']

#         print('')
#         print('Companion ', comp_ix, ':')
#         print('Separation', bestsep, 'pm', seperr)
#         print('Position angle', bestth, 'pm', therr)
#         print('Contrast ratio', bestcon, 'pm', conerr)

#     if plot:
#         p = pymultinest.PlotMarginalModes(a)
#         plt.figure(figsize=(5*n_params, 5*n_params))
#         plt.clf()
#         for i in range(n_params):
#             plt.subplot(n_params, n_params, n_params * i + i + 1)
#             p.plot_marginal(i, with_ellipses=True,
#                             with_points=False, grid_points=20)
#             plt.ylabel("Probability")
#             plt.xlabel(parameters[i])

#             for j in range(i):
#                 plt.subplot(n_params, n_params, n_params * j + i + 1)
#                 #plt.subplots_adjust(left=0, bottom=0, right=0, top=0, wspace=0, hspace=0)
#                 p.plot_conditional(i, j, with_ellipses=False,
#                                    with_points=True, grid_points=20)
#                 plt.xlabel(parameters[i])
#                 plt.ylabel(parameters[j])

#     return [a, s]

# =========================================================================
